from django import template

register = template.Library()


@register.inclusion_tag("portal/partials/benefits.html", takes_context=True)
def benefits(context):

    button1_link = context["BENEFITS"]["first"]["button"]["link"]
    link1_external = button1_link.startswith("http")

    button2_link = context["BENEFITS"]["second"]["button"]["link"]
    link2_external = button2_link.startswith("http")

    button3_link = context["BENEFITS"]["third"]["button"]["link"]
    link3_external = button3_link.startswith("http")

    return {
        "image1": context["BENEFITS"]["first"]["image"],
        "title1": context["BENEFITS"]["first"]["title"],
        "text1": context["BENEFITS"]["first"]["text"],
        "button1_text": context["BENEFITS"]["first"]["button"]["text"],
        "button1_link": button1_link,
        "link1_external": link1_external,
        "image2": context["BENEFITS"]["second"]["image"],
        "title2": context["BENEFITS"]["second"]["title"],
        "text2": context["BENEFITS"]["second"]["text"],
        "button2_text": context["BENEFITS"]["second"]["button"]["text"],
        "button2_link": button2_link,
        "link2_external": link2_external,
        "image3": context["BENEFITS"]["third"]["image"],
        "title3": context["BENEFITS"]["third"]["title"],
        "text3": context["BENEFITS"]["third"]["text"],
        "button3_text": context["BENEFITS"]["third"]["button"]["text"],
        "button3_link": button3_link,
        "link3_external": link3_external,
    }



import re
from builtins import object

from common.models import School
from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django_countries.widgets import CountrySelectWidget


class OrganisationForm(forms.ModelForm):
    county = forms.ChoiceField(
        choices=[
            [None, "(select county)"],
            ["Aberdeen City", "Aberdeen City"],
            ["Aberdeenshire", "Aberdeenshire"],
            ["Angus", "Angus"],
            ["Argyll and Bute", "Argyll and Bute"],
            ["Bedfordshire", "Bedfordshire"],
            ["Belfast", "Belfast"],
            ["Belfast Greater", "Belfast Greater"],
            ["Berkshire", "Berkshire"],
            ["Blaenau Gwent", "Blaenau Gwent"],
            ["Bridgend", "Bridgend"],
            ["Buckinghamshire", "Buckinghamshire"],
            ["Caerphilly", "Caerphilly"],
            ["Cambridgeshire", "Cambridgeshire"],
            ["Cardiff", "Cardiff"],
            ["Carmarthenshire", "Carmarthenshire"],
            ["Ceredigion", "Ceredigion"],
            ["Channel Islands", "Channel Islands"],
            ["Cheshire", "Cheshire"],
            ["City of Edinburgh", "City of Edinburgh"],
            ["Clackmannanshire", "Clackmannanshire"],
            ["Conwy", "Conwy"],
            ["Cornwall", "Cornwall"],
            ["County Antrim", "County Antrim"],
            ["County Armagh", "County Armagh"],
            ["County Down", "County Down"],
            ["County Fermanagh", "County Fermanagh"],
            ["County Londonderry", "County Londonderry"],
            ["County Tyrone", "County Tyrone"],
            ["County of Bristol", "County of Bristol"],
            ["Cumbria", "Cumbria"],
            ["Denbighshire", "Denbighshire"],
            ["Derbyshire", "Derbyshire"],
            ["Devon", "Devon"],
            ["Dorset", "Dorset"],
            ["Dumfries and Galloway", "Dumfries and Galloway"],
            ["Dunbartonshire", "Dunbartonshire"],
            ["Dundee City", "Dundee City"],
            ["Durham", "Durham"],
            ["East Ayrshire", "East Ayrshire"],
            ["East Dunbartonshire", "East Dunbartonshire"],
            ["East Lothian", "East Lothian"],
            ["East Renfrewshire", "East Renfrewshire"],
            ["East Riding of Yorkshire", "East Riding of Yorkshire"],
            ["East Sussex", "East Sussex"],
            ["Essex", "Essex"],
            ["Falkirk", "Falkirk"],
            ["Fife", "Fife"],
            ["Flintshire", "Flintshire"],
            ["Glasgow City", "Glasgow City"],
            ["Gloucestershire", "Gloucestershire"],
            ["Greater London", "Greater London"],
            ["Greater Manchester", "Greater Manchester"],
            ["Guernsey Channel Islands", "Guernsey Channel Islands"],
            ["Gwynedd", "Gwynedd"],
            ["Hampshire", "Hampshire"],
            ["Hereford and Worcester", "Hereford and Worcester"],
            ["Herefordshire", "Herefordshire"],
            ["Hertfordshire", "Hertfordshire"],
            ["Highland", "Highland"],
            ["Inverclyde", "Inverclyde"],
            ["Inverness", "Inverness"],
            ["Isle of Anglesey", "Isle of Anglesey"],
            ["Isle of Barra", "Isle of Barra"],
            ["Isle of Man", "Isle of Man"],
            ["Isle of Wight", "Isle of Wight"],
            ["Jersey Channel Islands", "Jersey Channel Islands"],
            ["Kent", "Kent"],
            ["Lancashire", "Lancashire"],
            ["Leicestershire", "Leicestershire"],
            ["Lincolnshire", "Lincolnshire"],
            ["Merseyside", "Merseyside"],
            ["Merthyr Tydfil", "Merthyr Tydfil"],
            ["Midlothian", "Midlothian"],
            ["Monmouthshire", "Monmouthshire"],
            ["Moray", "Moray"],
            ["Neath Port Talbot", "Neath Port Talbot"],
            ["Newport", "Newport"],
            ["Norfolk", "Norfolk"],
            ["North Ayrshire", "North Ayrshire"],
            ["North Lanarkshire", "North Lanarkshire"],
            ["North Yorkshire", "North Yorkshire"],
            ["Northamptonshire", "Northamptonshire"],
            ["Northumberland", "Northumberland"],
            ["Nottinghamshire", "Nottinghamshire"],
            ["Orkney", "Orkney"],
            ["Orkney Islands", "Orkney Islands"],
            ["Oxfordshire", "Oxfordshire"],
            ["Pembrokeshire", "Pembrokeshire"],
            ["Perth and Kinross", "Perth and Kinross"],
            ["Powys", "Powys"],
            ["Renfrewshire", "Renfrewshire"],
            ["Rhondda Cynon Taff", "Rhondda Cynon Taff"],
            ["Rutland", "Rutland"],
            ["Scottish Borders", "Scottish Borders"],
            ["Shetland Islands", "Shetland Islands"],
            ["Shropshire", "Shropshire"],
            ["Somerset", "Somerset"],
            ["South Ayrshire", "South Ayrshire"],
            ["South Lanarkshire", "South Lanarkshire"],
            ["South Yorkshire", "South Yorkshire"],
            ["Staffordshire", "Staffordshire"],
            ["Stirling", "Stirling"],
            ["Suffolk", "Suffolk"],
            ["Surrey", "Surrey"],
            ["Swansea", "Swansea"],
            ["Torfaen", "Torfaen"],
            ["Tyne and Wear", "Tyne and Wear"],
            ["Vale of Glamorgan", "Vale of Glamorgan"],
            ["Warwickshire", "Warwickshire"],
            ["West Dunbart", "West Dunbart"],
            ["West Lothian", "West Lothian"],
            ["West Midlands", "West Midlands"],
            ["West Sussex", "West Sussex"],
            ["West Yorkshire", "West Yorkshire"],
            ["Western Isles", "Western Isles"],
            ["Wiltshire", "Wiltshire"],
            ["Worcestershire", "Worcestershire"],
            ["Wrexham", "Wrexham"],
        ],
        required=False,
        help_text="County (optional)",
    )

    class Meta(object):
        model = School
        fields = ["name", "country", "county"]
        widgets = {
            "name": forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Name of school or club"}),
            "country": CountrySelectWidget(layout="{widget}"),
        }
        help_texts = {
            "name": "Name of school or club",
            "country": "Country (optional)",
        }

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user", None)
        self.current_school = kwargs.pop("current_school", None)
        super(OrganisationForm, self).__init__(*args, **kwargs)

    def clean(self):
        name = self.cleaned_data.get("name", None)

        if name:
            try:
                school = School.objects.get(name=name)
            except ObjectDoesNotExist:
                return self.cleaned_data

            if not self.current_school or self.current_school.id != school.id:
                raise forms.ValidationError("There is already a school or club registered with that name")

        return self.cleaned_data

    def clean_name(self):
        name = self.cleaned_data.get("name", None)

        if re.match(re.compile("^[\w\.' ]+$"), name) is None:
            raise forms.ValidationError(
                "School names cannot contain special characters except full stops and apostrophes."
            )

        return name



from selenium.webdriver.common.by import By

from portal.tests.pageObjects.portal.play.play_base_page import PlayBasePage


class JoinSchoolOrClubPage(PlayBasePage):
    def __init__(self, browser):
        super(JoinSchoolOrClubPage, self).__init__(browser)

        assert self.on_correct_page("play_join_organisation_page")

    def join_a_school_or_club(self, access_code):
        self.browser.find_element(By.ID, "id_access_code").send_keys(access_code)
        self.browser.find_element(By.ID, "request_join_a_school_or_club_button").click()
        assert self.element_exists_by_css(".success")

        return self

    def join_a_school_or_club_failure(self, access_code):
        self.browser.find_element(By.ID, "id_access_code").send_keys(access_code)
        self.browser.find_element(By.ID, "request_join_a_school_or_club_button").click()
        assert not self.element_exists_by_css(".success")

        return self

    def has_join_request_failed(self, error):
        errors = self.browser.find_element(By.ID, "join_class_form").find_element(By.CLASS_NAME, "errorlist").text
        return error in errors

    def revoke_join_request(self):
        self.browser.find_element(By.NAME, "revoke_join_request").click()

        return self



import itertools
import re
from builtins import map, range, str

from common.helpers.emails import send_verification_email
from common.models import Student, Teacher, UserSession, stripStudentName
from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.models import User
from django_recaptcha.fields import ReCaptchaField
from django_recaptcha.widgets import ReCaptchaV2Invisible
from game.models import Episode, Worksheet

from portal.forms.error_messages import INVALID_LOGIN_MESSAGE
from portal.helpers.password import PasswordStrength, form_clean_password
from portal.helpers.ratelimit import clear_ratelimit_cache_for_user


class InvitedTeacherForm(forms.Form):
    prefix = "teacher_signup"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field_name, field in self.fields.items():
            field.widget.attrs["id"] = f"id_teacher_signup-{field_name}"

    teacher_password = forms.CharField(
        help_text="Enter a password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
    )
    teacher_confirm_password = forms.CharField(
        help_text="Repeat password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Repeat password"}),
    )

    consent_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=True)
    newsletter_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=False)

    def clean_teacher_password(self):
        return form_clean_password(self, "teacher_password", PasswordStrength.TEACHER)

    def clean(self):
        if any(self.errors):
            return

        password = self.cleaned_data.get("teacher_password", None)
        confirm_password = self.cleaned_data.get("teacher_confirm_password", None)

        check_passwords(password, confirm_password)

        return self.cleaned_data


class TeacherSignupForm(InvitedTeacherForm):
    teacher_first_name = forms.CharField(
        help_text="Enter your first name",
        max_length=100,
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "First name"}),
    )
    teacher_last_name = forms.CharField(
        help_text="Enter your last name",
        max_length=100,
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Last name"}),
    )
    teacher_email = forms.EmailField(
        help_text="Enter your email address",
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
    )

    captcha = ReCaptchaField(widget=ReCaptchaV2Invisible)


class TeacherEditAccountForm(forms.Form):
    first_name = forms.CharField(
        max_length=100,
        widget=forms.TextInput(attrs={"placeholder": "First name", "class": "fName"}),
        help_text="First name",
    )
    last_name = forms.CharField(
        max_length=100,
        widget=forms.TextInput(attrs={"placeholder": "Last name", "class": "lName"}),
        help_text="Last name",
    )
    email = forms.EmailField(
        required=False,
        widget=forms.EmailInput(attrs={"placeholder": "New email address (optional)"}),
        help_text="New email address (optional)",
    )
    password = forms.CharField(
        required=False,
        widget=forms.PasswordInput(attrs={"placeholder": "New password (optional)"}),
        help_text="New password (optional)",
    )
    confirm_password = forms.CharField(
        required=False,
        widget=forms.PasswordInput(attrs={"placeholder": "Confirm new password (optional)"}),
        help_text="Confirm new password (optional)",
    )
    current_password = forms.CharField(
        widget=forms.PasswordInput(attrs={"placeholder": "Current password"}),
        help_text="Enter your current password",
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(TeacherEditAccountForm, self).__init__(*args, **kwargs)

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.TEACHER)

    def clean(self):
        if any(self.errors):
            return

        password = self.cleaned_data.get("password", None)
        confirm_password = self.cleaned_data.get("confirm_password", None)
        current_password = self.cleaned_data.get("current_password", None)

        self.check_password_errors(password, confirm_password, current_password)

        return self.cleaned_data

    def check_password_errors(self, password, confirm_password, current_password):
        check_passwords(password, confirm_password)

        if not self.user.check_password(current_password):
            raise forms.ValidationError("Your current password was incorrect")


class TeacherLoginForm(AuthenticationForm):
    username = forms.EmailField(
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
        help_text="Enter your email address",
    )
    password = forms.CharField(
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
        help_text="Enter your password",
    )

    def clean(self):
        email = self.cleaned_data.get("username", None)
        password = self.cleaned_data.get("password", None)

        if email and password:
            # Check it's a teacher and not a student using the same email address
            user = None

            user = self.find_user(email, user)

            user = authenticate(username=user.username, password=password)

            self.check_email_errors(user)

            self.user_cache = user

            # Reset ratelimit cache upon successful login
            clear_ratelimit_cache_for_user(user.username)

            # Log the login time and school
            teacher = Teacher.objects.get(new_user=user)
            session = UserSession(user=user, school=teacher.school)
            session.save()

        return self.cleaned_data

    def find_user(self, email, user):
        users = User.objects.filter(email=email)

        for result in users:
            if hasattr(result, "userprofile") and hasattr(result.userprofile, "teacher"):
                user = result
                break

        if user is None:
            self.show_invalid_login_message()

        return user

    def check_email_errors(self, user):
        if user is None:
            self.show_invalid_login_message()

        if not user.userprofile.is_verified:
            send_verification_email(self.request, user, self.data)
            self.show_invalid_login_message()

        if not user.is_active:
            raise forms.ValidationError("User account has been deactivated")

    def show_invalid_login_message(self):
        raise forms.ValidationError(INVALID_LOGIN_MESSAGE)


class ClassCreationForm(forms.Form):
    class_name = forms.CharField(
        widget=forms.TextInput(attrs={"placeholder": "Class name"}),
        help_text="Enter a class name",
    )
    teacher = forms.ChoiceField(help_text="Select teacher", required=False)
    classmate_progress = forms.BooleanField(
        label="Allow students to see their classmates' progress?",
        widget=forms.CheckboxInput(),
        initial=False,
        required=False,
    )

    def __init__(self, *args, teacher=None, **kwargs):
        super().__init__(*args, **kwargs)

        if teacher is not None:
            # Place current teacher at the top
            teacher_choices = [
                (
                    teacher.id,
                    f"{teacher.new_user.first_name} {teacher.new_user.last_name}",
                )
            ]

            # Get coworkers and add them to the choices if the teacher is an admin
            if teacher.is_admin:
                coworkers = (
                    Teacher.objects.filter(school=teacher.school)
                    .exclude(id=teacher.id)
                    .order_by("new_user__last_name", "new_user__first_name")
                )
                for coworker in coworkers:
                    teacher_choices.append(
                        (
                            coworker.id,
                            f"{coworker.new_user.first_name} {coworker.new_user.last_name}",
                        )
                    )

            self.fields["teacher"].choices = teacher_choices

    def clean(self):
        name = self.cleaned_data.get("class_name", "")

        if re.match(re.compile("^[\w -]+$"), name) is None:
            raise forms.ValidationError(
                "Class name may only contain letters, numbers, dashes, underscores, and spaces."
            )

        return self.cleaned_data


class ClassEditForm(forms.Form):
    # select dropdown choices for potentially limiting time in which external students may join
    # class
    # 0 value = don't allow
    # n value = allow for next n hours, n < 1000 hours
    # o/w = allow forever
    join_choices = [
        ("", "Don't change my current setting"),
        ("0", "Don't allow external requests to this class"),
        ("1", "Allow external requests to this class for the next hour"),
    ]
    join_choices.extend(
        [
            (
                str(hours),
                "Allow external requests to this class for the next " + str(hours) + " hours",
            )
            for hours in range(4, 28, 4)
        ]
    )
    join_choices.extend(
        [
            (
                str(days * 24),
                "Allow external requests to this class for the next " + str(days) + " days",
            )
            for days in range(2, 5)
        ]
    )
    join_choices.append(
        (
            "1000",
            "Always allow external requests to this class (not recommended)",
        )
    )
    name = forms.CharField(
        widget=forms.TextInput(attrs={"placeholder": "Enter class name"}),
        help_text="Enter class name",
    )
    classmate_progress = forms.BooleanField(
        label="Allow students to see their classmates' progress?",
        widget=forms.CheckboxInput(),
        initial=False,
        required=False,
    )
    external_requests = forms.ChoiceField(
        label="Set up external requests to this class",
        help_text="Choose your setting",
        required=False,
        choices=join_choices,
        widget=forms.Select(),
    )

    def clean(self):
        name = self.cleaned_data.get("name", "")

        if re.match(re.compile("^[\w -]+$"), name) is None:
            raise forms.ValidationError(
                "Class name may only contain letters, numbers, dashes, underscores, and spaces."
            )

        return self.cleaned_data


class ClassLevelControlForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super(ClassLevelControlForm, self).__init__(*args, **kwargs)

        episodes = Episode.objects.filter(pk__in=range(1, 25))

        for episode in episodes:
            choices = []
            for level in episode.levels:
                try:
                    choices.append((f"worksheet:{level.after_worksheet.id}", episode.name))
                except Worksheet.DoesNotExist:
                    pass
                choices.append((f"level:{level.id}", level.name))

            for worksheet in episode.worksheets.filter(before_level__isnull=True):
                choices.append((f"worksheet:{worksheet.id}", episode.name))

            self.fields[f"episode{episode.id}"] = forms.MultipleChoiceField(
                choices=itertools.chain(choices),
                widget=forms.CheckboxSelectMultiple(),
                required=False,
            )


class ClassMoveForm(forms.Form):
    new_teacher = forms.ChoiceField(
        label="New teacher to take over class",
        help_text="Select teacher",
        widget=forms.Select(),
    )

    def __init__(self, teachers, *args, **kwargs):
        self.teachers = teachers
        teacher_choices = []
        for teacher in teachers:
            teacher_choices.append(
                (
                    teacher.id,
                    teacher.new_user.first_name + " " + teacher.new_user.last_name,
                )
            )
        super(ClassMoveForm, self).__init__(*args, **kwargs)
        self.fields["new_teacher"].choices = teacher_choices


class TeacherEditStudentForm(forms.Form):
    name = forms.CharField(
        label="Name",
        widget=forms.TextInput(attrs={"placeholder": "Name"}),
        help_text="Choose a name",
    )

    def __init__(self, student, *args, **kwargs):
        self.student = student
        self.klass = student.class_field
        super(TeacherEditStudentForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        name = stripStudentName(self.cleaned_data.get("name", ""))

        if name == "":
            raise forms.ValidationError("'" + self.cleaned_data.get("name", "") + "' is not a valid name")

        if re.match(re.compile("^[\w -]+$"), name) is None:
            raise forms.ValidationError("Names may only contain letters, numbers, dashes, underscores, and spaces.")

        students = Student.objects.filter(class_field=self.klass, new_user__first_name__iexact=name)
        if students.exists() and students[0] != self.student:
            raise forms.ValidationError("There is already a student called '" + name + "' in this class")

        return name


class TeacherSetStudentPass(forms.Form):
    password = forms.CharField(
        label="New password",
        help_text="Enter new password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Enter new password"}),
    )
    confirm_password = forms.CharField(
        label="Confirm new password",
        help_text="Confirm new password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Confirm new password"}),
    )

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.STUDENT)

    def clean(self):
        password = self.cleaned_data.get("password", None)
        confirm_password = self.cleaned_data.get("confirm_password", None)

        # Student password is case insensitive
        if password is not None:
            password = password.lower()

        if confirm_password is not None:
            confirm_password = confirm_password.lower()

        check_passwords(password, confirm_password)

        return self.cleaned_data


def validateStudentNames(klass, names):
    validationErrors = []

    if klass:
        # We want to report if a student already exists with that name.
        # But only report each name once if there are duplicates.
        students = Student.objects.filter(class_field=klass)
        clashes_found = []
        find_clashes(names, students, clashes_found, validationErrors)

    # Also report if a student appears twice in the list to be added.
    # But again only report each name once.
    lower_names = [name.lower() for name in names]
    find_duplicates(names, lower_names, validationErrors)

    find_illegal_characters(names, validationErrors)

    return validationErrors


def find_clashes(names, students, clashes_found, validationErrors):
    for name in names:
        if students.filter(new_user__first_name__iexact=name).exists() and name not in clashes_found:
            validationErrors.append(
                forms.ValidationError("There is already a student called '" + name + "' in this class")
            )
            clashes_found.append(name)


def find_duplicates(names, lower_names, validationErrors):
    duplicates_found = []
    for duplicate in [name for name in names if lower_names.count(name.lower()) > 1]:
        if duplicate not in duplicates_found:
            validationErrors.append(
                forms.ValidationError("You cannot add more than one student called '" + duplicate + "'")
            )
            duplicates_found.append(duplicate)


def find_illegal_characters(names, validationErrors):
    for name in names:
        if re.match(re.compile("^[\w -]+$"), name) is None:
            validationErrors.append(
                forms.ValidationError(
                    "Names may only contain letters, numbers, dashes, underscores, and spaces. You must rename '"
                    + name
                    + "'."
                )
            )


def check_passwords(password, confirm_password):
    if password is not None and password != confirm_password:
        raise forms.ValidationError("The password and the confirmation password do not match")


class TeacherMoveStudentsDestinationForm(forms.Form):
    new_class = forms.ChoiceField(
        label="Choose a new class from the drop down menu for the student(s).",
        widget=forms.Select(attrs={"class": "wide"}),
    )

    def __init__(self, classes, *args, **kwargs):
        self.classes = classes
        class_choices = []
        for klass in classes:
            class_choices.append(
                (
                    klass.id,
                    klass.name
                    + " ("
                    + klass.access_code
                    + "), "
                    + klass.teacher.new_user.first_name
                    + " "
                    + klass.teacher.new_user.last_name,
                )
            )
        super(TeacherMoveStudentsDestinationForm, self).__init__(*args, **kwargs)
        self.fields["new_class"].choices = class_choices


class TeacherMoveStudentDisambiguationForm(forms.Form):
    orig_name = forms.CharField(
        label="Original Name",
        widget=forms.TextInput(
            attrs={
                "readonly": "readonly",
                "placeholder": "Original Name",
                "type": "hidden",
            }
        ),
    )
    name = forms.CharField(
        label="Name",
        widget=forms.TextInput(attrs={"placeholder": "Name", "style": "margin : 0px"}),
    )

    def clean_name(self):
        name = stripStudentName(self.cleaned_data.get("name", ""))
        if name == "":
            raise forms.ValidationError("'" + self.cleaned_data.get("name", "") + "' is not a valid name")
        return name


class BaseTeacherMoveStudentsDisambiguationFormSet(forms.BaseFormSet):
    def __init__(self, destination, *args, **kwargs):
        self.destination = destination
        super(BaseTeacherMoveStudentsDisambiguationFormSet, self).__init__(*args, **kwargs)

    def clean(self):
        if any(self.errors):
            return

        names = [form.cleaned_data["name"] for form in self.forms]

        validationErrors = validateStudentNames(self.destination, names)

        if len(validationErrors) > 0:
            raise forms.ValidationError(validationErrors)

        self.strippedNames = names


class TeacherDismissStudentsForm(forms.Form):
    orig_name = forms.CharField(
        help_text="Original student name",
        widget=forms.TextInput(
            attrs={
                "readonly": "readonly",
                "placeholder": "Original Name",
                "class": "m-0",
            }
        ),
    )
    name = forms.CharField(
        help_text="New student name",
        widget=forms.TextInput(attrs={"placeholder": "Enter new student name", "class": "m-0"}),
    )
    email = forms.EmailField(
        label="Email",
        help_text="New email address",
        widget=forms.EmailInput(attrs={"placeholder": "Enter email address", "class": "m-0"}),
    )
    confirm_email = forms.EmailField(
        label="Confirm Email",
        help_text="Confirm email address",
        widget=forms.EmailInput(attrs={"placeholder": "Confirm email address", "class": "m-0"}),
    )

    def clean_name(self):
        name = stripStudentName(self.cleaned_data.get("name", ""))

        if name == "":
            raise forms.ValidationError("'" + self.cleaned_data.get("name", "") + "' is not a valid name")

        return name

    def clean(self):
        email = self.cleaned_data.get("email", None)
        confirm_email = self.cleaned_data.get("confirm_email", None)

        if (email or confirm_email) and email != confirm_email:
            raise forms.ValidationError("Your new emails do not match")

        return self.cleaned_data


class BaseTeacherDismissStudentsFormSet(forms.BaseFormSet):
    def clean(self):
        if any(self.errors):
            return

        names = [form.cleaned_data["name"] for form in self.forms]

        validationErrors = validateStudentNames(None, names)

        if len(validationErrors) > 0:
            raise forms.ValidationError(validationErrors)


class StudentCreationForm(forms.Form):
    names = forms.CharField(
        label="names",
        widget=forms.Textarea(
            attrs={
                "placeholder": "You can import names from a .CSV file, or copy and paste them from a spreadsheet directly into this text box",
                "class": "m-0",
            }
        ),
    )

    def __init__(self, klass, *args, **kwargs):
        self.klass = klass
        super(StudentCreationForm, self).__init__(*args, **kwargs)

    def clean(self):
        names = re.split(";|,|\n", self.cleaned_data.get("names", ""))
        names = list(map(stripStudentName, names))
        names = [name for name in names if name != ""]

        validationErrors = validateStudentNames(self.klass, names)

        if len(validationErrors) > 0:
            raise forms.ValidationError(validationErrors)

        self.strippedNames = names

        return self.cleaned_data


class TeacherAddExternalStudentForm(forms.Form):
    name = forms.CharField(
        label="Student name",
        widget=forms.TextInput(attrs={"placeholder": "Name"}),
    )

    def __init__(self, klass, *args, **kwargs):
        self.klass = klass
        super(TeacherAddExternalStudentForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        name = stripStudentName(self.cleaned_data.get("name", ""))

        if name == "":
            raise forms.ValidationError("'" + self.cleaned_data.get("name", "") + "' is not a valid name")

        if Student.objects.filter(class_field=self.klass, new_user__first_name__iexact=name).exists():
            raise forms.ValidationError("There is already a student called '" + name + "' in this class")

        return name



import pytest


@pytest.mark.django_db
def test_preview_user_field_added(migrator):
    migrator.apply_initial_migration(("portal", "0054_pending_join_request_can_be_blank"))
    new_state = migrator.apply_tested_migration(("portal", "0055_add_preview_user"))

    userprofile_model = new_state.apps.get_model("portal", "UserProfile")
    assert userprofile_model._meta.get_field("preview_user").get_internal_type(), "BooleanField"

    school_model = new_state.apps.get_model("portal", "School")
    assert school_model._meta.get_field("eligible_for_testing").get_internal_type(), "BooleanField"



from django.contrib.auth.models import User
from common.models import Student

from common.helpers.generators import get_hashed_login_id


class StudentLoginBackend:
    def get_user(self, user_id):
        try:
            return User.objects.get(id=user_id)
        except User.DoesNotExist:
            return None

    def authenticate(self, request, user_id=None, login_id=None):
        """Check the credentials and return a user."""
        # Get the student by the user id
        user = self.get_user(user_id)
        if user:
            student = Student.objects.get(new_user=user)
            # Check the url against the student's stored hash then return the user.
            if student.login_id and get_hashed_login_id(login_id) == student.login_id:
                return user
        return None



from django.urls import reverse

from portal.tests.pageObjects.portal.base_page import BasePage
from portal.tests.pageObjects.portal.forbidden_page import ForbiddenPage


class AdminBasePage(BasePage):
    def __init__(self, browser, live_server_url):
        super(AdminBasePage, self).__init__(browser)
        self.live_server_url = live_server_url



from common.helpers.emails import (
    NOTIFICATION_EMAIL,
    delete_contact,
    send_email,
    update_indy_email,
)
from common.models import Student
from common.permissions import logged_in_as_student
from django.contrib import messages as messages
from django.contrib.auth import logout
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse_lazy
from django.views.generic.edit import FormView

from portal.forms.play import IndependentStudentEditAccountForm, StudentEditAccountForm
from portal.forms.registration import DeleteAccountForm
from portal.helpers.password import check_update_password
from portal.helpers.ratelimit import clear_ratelimit_cache_for_user
from portal.views.api import anonymise


def _get_form(self, form_class):
    """
    Generic function which gets the appropriate edit account details form for either
    a school student or an independent student.
    :param self: The view class the form is used in.
    :param form_class: The form class which corresponds to the type of student.
    :return: The initialised form which can then be used in the FormView.
    """
    user = self.request.user
    if form_class is None:
        form_class = self.get_form_class()
    return form_class(user, **self.get_form_kwargs())


def _process_form(self, process_function, form, view):
    """
    Generic function which processes the appropriate edit account details form for
    either a school student or an independent student.
    :param self: The view class the form is used in.
    :param process_function: The core function that is needed to process the form.
    :param form: The form which needs to be processed.
    :param view: The view object.
    :return: The view once the form is valid and has been submitted.
    """
    student = self.request.user.new_student
    process_function(form, student, self.request)
    return super(view, self).form_valid(form)


class SchoolStudentEditAccountView(LoginRequiredMixin, FormView):
    """
    A FormView for editing a school student's account details. This forms enables a
    school student to change their password.
    """

    login_url = reverse_lazy("student_login_access_code")
    form_class = StudentEditAccountForm
    template_name = "../templates/portal/play/student_edit_account.html"
    success_url = reverse_lazy("student_details")
    model = Student

    def form_valid(self, form):
        return _process_form(
            self,
            self.process_student_edit_account_form,
            form,
            SchoolStudentEditAccountView,
        )

    def process_student_edit_account_form(self, form, student, request):
        data = form.cleaned_data
        # check not default value for CharField
        changing_password = check_update_password(form, student.new_user, request, data)

        messages.success(request, "Your account details have been changed successfully.")

        if changing_password:
            logout(request)
            messages.success(
                request,
                "Please login using your new password.",
            )
            return HttpResponseRedirect(reverse_lazy("student_login_access_code"))

    def get_form(self, form_class=None):
        return _get_form(self, form_class)


def independentStudentEditAccountView(request):
    """
    A FormView for editing an independent student's account details. This forms enables
    an independent student to change their name, their email and / or their password.
    Additionally it also deletes the independent students account with its second form

    Django class based views makes it very difficult to have multiple forms in one view,
    hence using the functional based view
    """
    user = request.user
    change_email_password_form = IndependentStudentEditAccountForm(user, initial={"name": user.first_name})
    delete_account_form = DeleteAccountForm(user)
    template_name = "../templates/portal/play/student_edit_account.html"
    delete_account_confirm = False

    def process_change_email_password_form(request, student, old_anchor):
        change_email_password_form = IndependentStudentEditAccountForm(request.user, request.POST)
        changing_email = False
        changing_password = False
        new_email = ""
        if change_email_password_form.is_valid():
            data = change_email_password_form.cleaned_data
            # check not default value for CharField
            changing_password = check_update_password(change_email_password_form, student, request, data)

            changing_email, new_email = update_indy_email(student, request, data)
            changing_first_name = False if request.POST["name"] == student.first_name else True
            student.first_name = request.POST["name"]

            student.save()

            anchor = ""

            # Reset ratelimit cache after successful account details update
            clear_ratelimit_cache_for_user(student.username)

            messages.success(request, "Your account details have been successfully changed.")
        else:
            anchor = old_anchor

        return changing_email, new_email, changing_password, changing_first_name, anchor

    if request.method == "POST":
        if "delete_account" in request.POST:
            delete_account_form = DeleteAccountForm(user, request.POST)
            if not delete_account_form.is_valid():
                messages.warning(request, "Your account was not deleted due to incorrect password.")
            else:
                delete_account_confirm = True
        else:
            change_email_password_form = IndependentStudentEditAccountForm(request.user, request.POST)
            if not change_email_password_form.is_valid():
                messages.warning(request, "Your account was not updated do to incorrect details")
                return render(
                    request,
                    template_name,
                    {
                        "form": change_email_password_form,
                        "delete_account_form": delete_account_form,
                    },
                )
            else:
                (
                    changing_email,
                    new_email,
                    changing_password,
                    changing_first_name,
                    anchor,
                ) = process_change_email_password_form(request, request.user, "")
                if changing_email:
                    logout(request)
                    messages.success(request, "Your account details have been changed successfully.")
                    messages.success(
                        request,
                        "Your email will be changed once you have verified it, until then "
                        "you can still log in with your old email.",
                    )
                    return render(
                        request,
                        "portal/email_verification_needed.html",
                        {"usertype": "INDEP_STUDENT"},
                    )
                if changing_password:
                    logout(request)
                    messages.success(request, "Your account details have been changed successfully.")
                    messages.success(request, "Please login using your new password.")
                    return HttpResponseRedirect(reverse_lazy("independent_student_login"))

                if changing_first_name:
                    messages.success(request, "Your account details have been changed successfully.")
                    return HttpResponseRedirect(reverse_lazy("independent_student_details"))

    return render(
        request,
        template_name,
        {
            "form": change_email_password_form,
            "delete_account_form": delete_account_form,
            "delete_account_confirm": delete_account_confirm,
        },
    )


@login_required(login_url=reverse_lazy("home"))
@user_passes_test(logged_in_as_student, login_url=reverse_lazy("home"))
def student_edit_account(request):
    student = request.user.new_student
    if student.is_independent():
        return HttpResponseRedirect(reverse_lazy("independent_edit_account"))
    else:
        return HttpResponseRedirect(reverse_lazy("school_student_edit_account"))



from selenium.webdriver.common.by import By

from portal.tests.pageObjects.portal.teach.added_independent_student_to_class_page import (
    AddedIndependentStudentToClassPage,
)
from portal.tests.pageObjects.portal.teach.teach_base_page import TeachBasePage


class AddIndependentStudentToClassPage(TeachBasePage):
    def __init__(self, browser):
        super(AddIndependentStudentToClassPage, self).__init__(browser)

        assert self.on_correct_page("add_independent_student_to_class")

    def _rename(self, name):
        self.browser.find_element(By.ID, "id_name").clear()
        self.browser.find_element(By.ID, "id_name").send_keys(name)

    def save(self, name):
        self._rename(name)
        self.wait_for_element_to_be_clickable((By.ID, "save_student_name_button"))
        self.browser.find_element(By.ID, "save_student_name_button").click()

        return AddedIndependentStudentToClassPage(self.browser)



from pathlib import Path

from selenium.webdriver.common.by import By

from portal.tests.pageObjects.portal.base_page import BasePage


class TeachBasePage(BasePage):
    def __init__(self, browser):
        super(TeachBasePage, self).__init__(browser)

    def logout(self):
        self.browser.find_element(By.ID, "logout_menu").click()

        self.browser.find_element(By.ID, "logout_button").click()
        from portal.tests.pageObjects.portal.home_page import HomePage

        return HomePage(self.browser)

    def import_students_from_csv(self, filename):
        self.browser.execute_script(
            """
            $(document).ready(function () {
                const fileInput = $('<input>').attr({
                    type: 'file'
                })
                fileInput.on('change', studentsCsvChange('#id_names'))
                $('body').append(fileInput);
            })
        """
        )
        self.browser.find_element(By.XPATH, "/html/body/input").send_keys(
            str((Path(__file__).parents[3] / "data" / filename).resolve())
        )

        return self

    def get_students_input_value(self):
        return self.browser.find_element(By.ID, "id_names").get_attribute("value")



import datetime
import uuid

from common.models import Class, School, Student, Teacher, UserProfile
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.http import HttpRequest, HttpResponse
from django.utils import timezone
from rest_framework import generics, permissions, serializers, status
from rest_framework.authentication import SessionAuthentication
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import reverse_lazy

from portal.app_settings import IS_CLOUD_SCHEDULER_FUNCTION

THREE_YEARS_IN_DAYS = 1095


@api_view(("GET",))
@login_required(login_url=reverse_lazy("administration_login"))
def registered_users(request, year, month, day):
    try:
        nbr_reg = User.objects.filter(
            date_joined__startswith=datetime.date(
                int(year), int(month), int(day)
            )
        ).count()
        return Response(nbr_reg)
    except ValueError:
        return HttpResponse(status=404)


@api_view(("GET",))
@login_required(login_url=reverse_lazy("administration_login"))
def last_connected_since(request, year, month, day):
    try:
        nbr_active_users = User.objects.filter(
            last_login__gte=datetime.date(int(year), int(month), int(day))
        ).count()
        return Response(nbr_active_users)
    except ValueError:
        return HttpResponse(status=404)


@api_view(("GET",))
@login_required(login_url=reverse_lazy("administration_login"))
def number_users_per_country(request, country):
    try:
        nbr_reg = (
            Teacher.objects.filter(school__country__exact=country).count()
            + Student.objects.filter(
                class_field__teacher__school__country__exact=country
            ).count()
        )
        return Response(nbr_reg)
    except ValueError:
        return HttpResponse(status=404)


class InactiveUserSerializer(serializers.Serializer):
    """The user information we show in the InactiveUsersViewSet."""

    username = serializers.CharField()
    first_name = serializers.CharField()
    last_name = serializers.CharField()
    email = serializers.CharField()
    date_joined = serializers.DateTimeField()
    last_login = serializers.DateTimeField()


class IsAdminOrGoogleAppEngine(permissions.IsAdminUser):
    """Checks whether the request is from a Google App Engine cron job."""

    def has_permission(self, request: HttpRequest, view):
        is_admin = super(IsAdminOrGoogleAppEngine, self).has_permission(
            request, view
        )
        return IS_CLOUD_SCHEDULER_FUNCTION(request) or is_admin


def __anonymise_user(user):
    # the actual user anonymisation
    user.username = uuid.uuid4().hex
    user.first_name = "Deleted"
    user.last_name = "User"
    user.email = ""
    user.is_active = False
    user.save()


def anonymise(user):
    """Anonymise user. If admin teacher, pass the admin role to another teacher (if exists).
    If the only teacher, anonymise the school.
    """
    is_admin = False
    teacher = None
    # Find the teacher even if they're anonymised
    teacher_set = Teacher._base_manager.filter(new_user=user)
    if teacher_set:
        is_admin = teacher_set[0].is_admin
        school = teacher_set[0].school
        teacher = teacher_set[0]

    __anonymise_user(user)

    # if teacher, anonymise classes and students
    if teacher:
        classes = Class.objects.filter(teacher=teacher)
        for klass in classes:
            students = Student.objects.filter(class_field=klass)
            for student in students:
                __anonymise_user(student.new_user)
            klass.anonymise()

    # if user is admin and the school does not have another admin, appoint another teacher as admin
    if is_admin:
        teachers = Teacher.objects.filter(school=school).order_by(
            "new_user__last_name", "new_user__first_name"
        )
        if not teachers:
            # no other teacher, anonymise the school
            school.anonymise()
            return

        admin_exists = False
        for teacher in teachers:
            if teacher.is_admin:
                admin_exists = True
                break

        # if no admin, appoint the first teacher as admin
        if not admin_exists:
            teachers[0].is_admin = True
            teachers[0].save()


class InactiveUsersView(generics.ListAPIView):
    """
    This API view endpoint allows us to see our inactive users.

    An inactive user is one that hasn't logged in for three years.
    If the user has never logged in, we look at the date they registered with us instead.
    """

    queryset = User.objects.filter(is_active=True) & (
        User.objects.filter(
            last_login__lte=timezone.now()
            - timezone.timedelta(days=THREE_YEARS_IN_DAYS)
        )
        | User.objects.filter(
            last_login__isnull=True,
            date_joined__lte=timezone.now()
            - timezone.timedelta(days=THREE_YEARS_IN_DAYS),
        )
    )
    authentication_classes = (SessionAuthentication,)
    serializer_class = InactiveUserSerializer
    permission_classes = (IsAdminOrGoogleAppEngine,)

    def delete(self, request: HttpRequest):
        """Delete all personal data from inactive users and mark them as inactive."""
        inactive_users = self.get_queryset()
        for user in inactive_users:
            anonymise(user)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RemoveFakeAccounts(generics.ListAPIView):
    """
    This API endpoint will delete suspicious accounts that have the same first and last name and who are not verified
    """

    authentication_classes = (SessionAuthentication,)
    serializer_class = InactiveUserSerializer
    permission_classes = (IsAdminOrGoogleAppEngine,)

    def get(self, request):
        userprofiles = UserProfile.objects.filter(is_verified=False)
        [
            userprofile.user.delete()
            for userprofile in userprofiles
            if userprofile.user.first_name == userprofile.user.last_name
        ]

        return HttpResponse(status=204)


class AnonymiseOrphanSchoolsView(generics.ListAPIView):
    authentication_classes = (SessionAuthentication,)
    serializer_class = InactiveUserSerializer
    permission_classes = (IsAdminOrGoogleAppEngine,)

    def get(self, request: HttpRequest, start_id):
        # Re-anonymise all inactive teachers so their schools (if necessary) and classes/students are anonymised
        for teacher in Teacher._base_manager.filter(
            pk__gte=start_id, new_user__is_active=False
        ):
            anonymise(teacher.new_user)

        # Anonymise schools without any teachers
        for school in School.objects.filter(teacher_school__isnull=True):
            school.anonymise()

        return Response(status=status.HTTP_204_NO_CONTENT)



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("portal", "0045_auto_20150430_1446")]

    operations = [
        migrations.AlterModelOptions(
            name="school",
            options={
                "permissions": (
                    ("view_aggregated_data", "Can see available aggregated data"),
                    ("view_map_data", "Can see schools' location displayed on map"),
                )
            },
        )
    ]



from __future__ import absolute_import

import json
import time
from unittest.mock import Mock, patch

import pytest
from common.models import JoinReleaseStudent
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import (
    create_organisation_directly,
    join_teacher_to_organisation,
)
from common.tests.utils.student import (
    create_many_school_students,
    create_school_student,
    create_school_student_directly,
    create_student_with_direct_login,
)
from common.tests.utils.teacher import signup_teacher_directly
from django.contrib.auth.models import User
from django.test import Client
from django.urls import reverse
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.by import By

from portal.tests.pageObjects.portal.home_page import HomePage
from .base_test import BaseTest


class TestTeacherStudentFrontend(BaseTest):
    def test_create_valid_name(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        create_class_directly(email)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_students(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        student_name = "Florian-Gilbert"
        page = page.type_student_name(student_name)

        student_name2 = "Florian_Gilbert"
        page = page.type_student_name(student_name2)

        page.click_create_students()

        assert page.student_exists(student_name)
        assert page.student_exists(student_name2)

    def test_create_invalid_name(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        create_class_directly(email)

        student_name = "Florian!"

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_students(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        page = page.type_student_name(student_name).click_create_students()

        assert page.adding_students_failed()
        assert page.was_form_invalid(
            "form-create-students", "Names may only contain letters, numbers, dashes, underscores, and spaces."
        )

    def test_create_multiple(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        create_class_directly(email)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_students(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        page, student_names = create_many_school_students(page, 12)

        for student_name in student_names:
            assert page.student_exists(student_name)

    def test_create_duplicate(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        create_class_directly(email)

        student_name = "bob"

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_students(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        page = page.type_student_name(student_name).type_student_name(student_name).click_create_students()
        assert page.adding_students_failed()
        assert page.duplicate_students(student_name)

    def test_onboarding_import_from_csv(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        create_class_directly(email)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_students(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .import_students_from_csv("test_students_names.csv")
        )

        assert page.get_students_input_value() == "Student 1\nStudent 2\n"

    def test_onboarding_import_from_csv_error(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        create_class_directly(email)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login_no_students(email, password)
            .import_students_from_csv("test_students_names_no_name.csv")
        )

        time.sleep(1)

        alert = Alert(page.browser)
        assert alert.text == "'Name' column not found in CSV file."
        alert.dismiss()

    def test_class_students_import_from_csv(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .import_students_from_csv("test_students_names.csv")
        )

        assert page.get_students_input_value() == "Student 1\nStudent 2\n"

    def test_class_students_import_from_csv_error(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .import_students_from_csv("test_students_names_no_name.csv")
        )

        time.sleep(1)

        alert = Alert(page.browser)
        assert alert.text == "'Name' column not found in CSV file."
        alert.dismiss()

    def test_add_to_existing_class(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        page, new_student_name = create_school_student(page)
        assert page.student_exists(new_student_name)

        page = page.go_back_to_class()

        assert page.student_exists(new_student_name)

    def test_new_student_can_login_with_url(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        page, new_student_name = create_school_student(page)
        assert page.student_exists(new_student_name)

        # get login url, then open it and check if the student is logged in
        login_url = page.get_first_login_url()
        page.browser.get(login_url)
        assert page.on_correct_page("play_dashboard_page")
        assert new_student_name in page.browser.find_element(By.XPATH, "//div[@class='header']").text

    def test_update_student_name(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .go_to_edit_student_page()
        )

        assert page.is_student_name(name)

        new_student_name = "new name"

        page = page.type_student_name(new_student_name)
        page = page.click_update_button()

        assert page.__class__.__name__ == "TeachClassPage"
        assert page.student_exists(new_student_name)

    def test_update_student_valid_name_dash(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .go_to_edit_student_page()
        )

        assert page.is_student_name(name)

        new_student_name = "new-name"

        page = page.type_student_name(new_student_name)
        page = page.click_update_button()

        assert page.__class__.__name__ == "TeachClassPage"
        assert page.student_exists(new_student_name)

    def test_update_student_valid_name_underscore(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .go_to_edit_student_page()
        )

        assert page.is_student_name(name)

        new_student_name = "new_name"

        page = page.type_student_name(new_student_name)
        page = page.click_update_button()

        assert page.__class__.__name__ == "TeachClassPage"
        assert page.student_exists(new_student_name)

    def test_update_student_invalid_name(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .go_to_edit_student_page()
        )

        assert page.is_student_name(name)

        new_student_name = "new name!"

        page = page.type_student_name(new_student_name)
        page = page.click_update_button_fail()

        assert page.is_student_name(name)
        assert page.was_form_invalid(
            "form-edit-student", "Names may only contain letters, numbers, dashes, underscores, and spaces."
        )

    def test_update_student_password(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
            .go_to_edit_student_page()
        )

        assert page.is_student_name(name)

        new_student_password = "£EDCVFR$5tgb"

        page = page.type_student_password(new_student_password)
        page = page.click_set_password_button()

        assert page.student_exists(name)
        assert page.is_student_password(new_student_password.lower())

    def test_delete_student(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, _, student = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )
        assert page.student_exists(student_name)

        page = page.toggle_select_student().delete_students()
        assert page.is_dialog_showing()
        page = page.confirm_delete_student_dialog()

        assert not page.student_exists(student_name)

        # user/student is removed if they never log in (see below for active student)
        with pytest.raises(User.DoesNotExist):
            User.objects.get(id=student.new_user.id)

    def test_delete_active_student(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student, login_id, student_name, student_password = create_student_with_direct_login(access_code)

        # "active student" is one who has logged in
        c = Client()
        url = reverse("student_login", kwargs={"access_code": access_code})
        c.post(url, {"username": student_name, "password": student_password})

        # teacher login
        c.login(username=email, password=password)

        # delete the student
        url = reverse("teacher_delete_students", args=[access_code])
        data = {"transfer_students": json.dumps([student.id])}
        c.post(url, data)

        # user should be anonymised
        u = User.objects.get(id=student.new_user.id)
        assert u.first_name == "Deleted"
        assert not u.is_active

        student.refresh_from_db()
        assert student.login_id == ""

        # try to login as inactive student, it should fail
        url = f"/u/{student.user.id}/{login_id}/"
        response = c.get(url)

        # assert redirects to home page (failed to login)
        assert response.status_code == 302
        assert response.url == "/"

    def test_reset_passwords(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name, _, _ = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )
        assert page.student_exists(student_name)

        page = page.toggle_select_student().reset_passwords()
        assert page.is_dialog_showing()
        page = page.confirm_reset_student_dialog()

        assert page.student_exists(student_name)
        assert page.__class__.__name__ == "OnboardingStudentListPage"

    def test_move_cancel(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )

        page = page.move_students_none_selected()
        assert page.__class__.__name__ == "TeachClassPage"

        page = page.toggle_select_student().move_students()
        assert page.__class__.__name__ == "TeachMoveStudentsPage"

        page = page.cancel()
        assert page.__class__.__name__ == "TeachClassPage"

    def test_move_cancel_disambiguate(self):
        old_teacher_email, password_1 = signup_teacher_directly()
        email_2, _ = signup_teacher_directly()
        school = create_organisation_directly(old_teacher_email)
        join_teacher_to_organisation(email_2, school.name)
        _, _, access_code_1 = create_class_directly(old_teacher_email)
        create_class_directly(email_2)
        student_name, _, _ = create_school_student_directly(access_code_1)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(old_teacher_email, password_1)
            .open_classes_tab()
            .go_to_class_page()
        )
        assert page.has_students()
        assert page.student_exists(student_name)

        page = page.toggle_select_student()
        page = page.move_students().select_class_by_index(0).move().cancel()
        assert page.has_students()
        assert page.student_exists(student_name)

    def test_move(self):
        email_1, password_1 = signup_teacher_directly()
        email_2, password_2 = signup_teacher_directly()
        school = create_organisation_directly(email_1)
        join_teacher_to_organisation(email_2, school.name)
        _, _, access_code_1 = create_class_directly(email_1)
        create_class_directly(email_2)
        student_name_1, _, _ = create_school_student_directly(access_code_1)
        student_name_2, _, _ = create_school_student_directly(access_code_1)
        # Sort student names alphabetically to match the UI
        if student_name_1 > student_name_2:
            student_name_1, student_name_2 = student_name_2, student_name_1

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email_1, password_1)
            .open_classes_tab()
            .go_to_class_page()
        )
        assert page.student_exists(student_name_1)
        assert page.student_exists(student_name_2)

        page = page.toggle_select_student()
        page = page.move_students().select_class_by_index(0).move().move()
        assert not page.student_exists(student_name_1)

        page = page.go_to_dashboard()

        page = page.logout().go_to_teacher_login_page().login(email_2, password_2).open_classes_tab().go_to_class_page()
        assert page.student_exists(student_name_1)

    @patch("common.helpers.emails.send_dotdigital_email")
    def test_dismiss(self, mock_send_dotdigital_email: Mock):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name_1, _, student = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )
        assert page.student_exists(student_name_1)

        page = page.toggle_select_student().dismiss_students()
        assert page.__class__.__name__ == "TeachDismissStudentsPage"
        page = page.cancel()
        assert page.__class__.__name__ == "TeachClassPage"

        page = page.toggle_select_student().dismiss_students().enter_email("student_email@gmail.com").dismiss()
        assert not page.student_exists(student_name_1)

        # check whether a record is created correctly
        logs = JoinReleaseStudent.objects.filter(student=student)
        assert len(logs) == 1
        assert logs[0].action_type == JoinReleaseStudent.RELEASE

        mock_send_dotdigital_email.assert_called()

    @patch("common.helpers.emails.send_dotdigital_email")
    def test_multiple_dismiss(self, mock_send_dotdigital_email: Mock):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        student_name_1, _, student = create_school_student_directly(access_code)
        student_name_2, _, student_2 = create_school_student_directly(access_code)

        self.selenium.get(self.live_server_url)
        page = (
            HomePage(self.selenium)
            .go_to_teacher_login_page()
            .login(email, password)
            .open_classes_tab()
            .go_to_class_page()
        )
        assert page.student_exists(student_name_1)
        assert page.student_exists(student_name_2)

        page = page.toggle_all_students()
        page = page.dismiss_students()

        # dismiss with the same email address
        SAME_EMAIL = "student_email@gmail.com"
        page = page.enter_email(SAME_EMAIL, 0)
        page = page.enter_email(SAME_EMAIL, 1)
        page = page.dismiss()

        # the first should be released, the second not
        assert not page.student_exists(student_name_1)
        assert page.student_exists(student_name_2)

        # dismiss using teacher's email
        page = page.toggle_all_students()
        page = page.dismiss_students()

        page = page.enter_email(email, 0)
        page = page.dismiss()

        # student should still exist
        assert page.student_exists(student_name_2)

        mock_send_dotdigital_email.assert_called()



import logging
from datetime import datetime, timedelta

from common.helpers.emails import generate_token_for_email
from common.mail import campaign_ids, send_dotdigital_email
from common.models import DailyActivity, TotalActivity
from django.contrib.auth.models import User
from django.db.models import F, Q
from django.db.models.query import QuerySet
from django.urls import reverse
from django.utils import timezone
from rest_framework.response import Response
from rest_framework.views import APIView

from ...mixins import CronMixin
from ...views.api import anonymise

USER_1ST_VERIFY_EMAIL_REMINDER_DAYS = 7
USER_2ND_VERIFY_EMAIL_REMINDER_DAYS = 14
USER_DELETE_UNVERIFIED_ACCOUNT_DAYS = 19

USER_1ST_INACTIVE_REMINDER_DAYS = 730  # 2 years
USER_2ND_INACTIVE_REMINDER_DAYS = 973  # roughly 2 years and 8 months
USER_FINAL_INACTIVE_REMINDER_DAYS = 1065  # 2 years and 11 months
USER_RETENTION_PERIOD = 1096  # 3 years


def get_unverified_users(
    days: int, same_day: bool
) -> (QuerySet[User], QuerySet[User]):
    now = timezone.now()

    # All expired unverified users.
    user_queryset = User.objects.filter(
        date_joined__lte=now - timedelta(days=days),
        userprofile__is_verified=False,
    )
    if same_day:
        user_queryset = user_queryset.filter(
            date_joined__gt=now - timedelta(days=days + 1)
        )

    teacher_queryset = user_queryset.filter(
        new_teacher__isnull=False,
        new_student__isnull=True,
    )
    independent_student_queryset = user_queryset.filter(
        new_teacher__isnull=True,
        new_student__class_field__isnull=True,
    )

    return teacher_queryset, independent_student_queryset


def get_inactive_users(days: int) -> QuerySet[User]:
    now = timezone.now()

    # All users who haven't logged in in X days OR who've never logged in and
    # registered over X days ago.
    user_queryset = User.objects.filter(
        Q(
            last_login__isnull=False,
            last_login__lte=now - timedelta(days=days),
            last_login__gt=now - timedelta(days=days + 1),
        )
        | Q(
            last_login__isnull=True,
            date_joined__lte=now - timedelta(days=days),
            date_joined__gt=now - timedelta(days=days + 1),
        )
    )

    return user_queryset.exclude(email__isnull=True).exclude(email="")


def build_absolute_google_uri(request, location: str) -> str:
    """
    This is needed specifically for emails sent by cron jobs as the protocol for
    cron jobs is HTTP and the service name is wrongly parsed.
    """
    url = request.build_absolute_uri(location)
    url = url.replace("http", "https")
    url = url.replace(".decent", "-dot-decent")

    return url


class FirstVerifyEmailReminderView(CronMixin, APIView):
    def get(self, request):
        teacher_queryset, independent_student_queryset = get_unverified_users(
            USER_1ST_VERIFY_EMAIL_REMINDER_DAYS,
            same_day=True,
        )
        user_queryset = teacher_queryset.union(independent_student_queryset)
        user_count = user_queryset.count()

        logging.info(f"{user_count} emails unverified.")

        if user_count > 0:
            sent_email_count = 0
            for email in user_queryset.values_list("email", flat=True).iterator(
                chunk_size=500
            ):
                email_verification_url = build_absolute_google_uri(
                    request,
                    reverse(
                        "verify_email",
                        kwargs={"token": generate_token_for_email(email)},
                    ),
                )

                try:
                    send_dotdigital_email(
                        campaign_ids["verify_new_user_first_reminder"],
                        [email],
                        personalization_values={
                            "VERIFICATION_LINK": email_verification_url
                        },
                    )

                    sent_email_count += 1
                except Exception as ex:
                    logging.exception(ex)

            logging.info(f"Sent {sent_email_count}/{user_count} emails.")

        return Response()


class SecondVerifyEmailReminderView(CronMixin, APIView):
    def get(self, request):
        teacher_queryset, independent_student_queryset = get_unverified_users(
            USER_2ND_VERIFY_EMAIL_REMINDER_DAYS,
            same_day=True,
        )
        user_queryset = teacher_queryset.union(independent_student_queryset)
        user_count = user_queryset.count()

        logging.info(f"{user_count} emails unverified.")

        if user_count > 0:

            sent_email_count = 0
            for email in user_queryset.values_list("email", flat=True).iterator(
                chunk_size=500
            ):
                email_verification_url = build_absolute_google_uri(
                    request,
                    reverse(
                        "verify_email",
                        kwargs={"token": generate_token_for_email(email)},
                    ),
                )

                try:
                    send_dotdigital_email(
                        campaign_ids["verify_new_user_second_reminder"],
                        [email],
                        personalization_values={
                            "VERIFICATION_LINK": email_verification_url
                        },
                    )

                    sent_email_count += 1
                except Exception as ex:
                    logging.exception(ex)

            logging.info(f"Sent {sent_email_count}/{user_count} emails.")

        return Response()


class AnonymiseUnverifiedAccounts(CronMixin, APIView):
    def get(self, request):
        user_count = User.objects.filter(is_active=True).count()

        teacher_queryset, independent_student_queryset = get_unverified_users(
            USER_DELETE_UNVERIFIED_ACCOUNT_DAYS,
            same_day=False,
        )
        teacher_count = teacher_queryset.count()
        indy_count = independent_student_queryset.count()

        user_queryset = teacher_queryset.union(independent_student_queryset)

        for user in user_queryset.iterator(chunk_size=100):
            try:
                anonymise(user)
            except Exception as ex:
                logging.error(f"Failed to anonymise user with id: {user.id}")
                logging.exception(ex)

        user_count -= User.objects.filter(is_active=True).count()
        logging.info(f"{user_count} unverified users anonymised.")

        activity_today = DailyActivity.objects.get_or_create(
            date=datetime.now().date()
        )[0]
        activity_today.anonymised_unverified_teachers = teacher_count
        activity_today.anonymised_unverified_independents = indy_count
        activity_today.save()

        TotalActivity.objects.update(
            anonymised_unverified_teachers=F("anonymised_unverified_teachers")
            + teacher_count,
            anonymised_unverified_independents=F(
                "anonymised_unverified_independents"
            )
            + indy_count,
        )

        return Response()


class FirstInactivityReminderView(CronMixin, APIView):
    def get(self, request):
        user_queryset = get_inactive_users(USER_1ST_INACTIVE_REMINDER_DAYS)
        user_count = user_queryset.count()

        logging.info(
            f"{user_count} inactive users after "
            f"{USER_1ST_INACTIVE_REMINDER_DAYS} days."
        )

        if user_count > 0:
            sent_email_count = 0
            for email in user_queryset.values_list("email", flat=True).iterator(
                chunk_size=500
            ):
                try:
                    send_dotdigital_email(
                        campaign_ids[
                            "inactive_users_on_website_first_reminder"
                        ],
                        [email],
                    )

                    sent_email_count += 1
                except Exception as ex:
                    logging.exception(ex)

            logging.info(
                f"Reminded {sent_email_count}/{user_count} inactive users."
            )

        return Response()


class SecondInactivityReminderView(CronMixin, APIView):
    def get(self, request):
        user_queryset = get_inactive_users(USER_2ND_INACTIVE_REMINDER_DAYS)
        user_count = user_queryset.count()

        logging.info(
            f"{user_count} inactive users after "
            f"{USER_2ND_INACTIVE_REMINDER_DAYS} days."
        )

        if user_count > 0:
            sent_email_count = 0
            for email in user_queryset.values_list("email", flat=True).iterator(
                chunk_size=500
            ):
                try:
                    send_dotdigital_email(
                        campaign_ids[
                            "inactive_users_on_website_second_reminder"
                        ],
                        [email],
                        personalization_values={
                            "DAYS_LEFT": USER_RETENTION_PERIOD
                            - USER_2ND_INACTIVE_REMINDER_DAYS
                        },
                    )

                    sent_email_count += 1
                except Exception as ex:
                    logging.exception(ex)

            logging.info(
                f"Reminded {sent_email_count}/{user_count} inactive users."
            )

        return Response()


class FinalInactivityReminderView(CronMixin, APIView):
    def get(self, request):
        user_queryset = get_inactive_users(USER_FINAL_INACTIVE_REMINDER_DAYS)
        user_count = user_queryset.count()

        logging.info(
            f"{user_count} inactive users after "
            f"{USER_FINAL_INACTIVE_REMINDER_DAYS} days."
        )

        if user_count > 0:
            sent_email_count = 0
            for email in user_queryset.values_list("email", flat=True).iterator(
                chunk_size=500
            ):
                try:
                    send_dotdigital_email(
                        campaign_ids[
                            "inactive_users_on_website_final_reminder"
                        ],
                        [email],
                        personalization_values={
                            "DAYS_LEFT": USER_RETENTION_PERIOD
                            - USER_FINAL_INACTIVE_REMINDER_DAYS
                        },
                    )

                    sent_email_count += 1
                except Exception as ex:
                    logging.exception(ex)

            logging.info(
                f"Reminded {sent_email_count}/{user_count} inactive users."
            )

        return Response()



from __future__ import absolute_import

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select

from ..email_verification_needed_page import EmailVerificationNeededPage
from .add_independent_student_to_class_page import \
    AddIndependentStudentToClassPage
from .class_page import TeachClassPage
from .move_classes_page import TeachMoveClassesPage
from .teach_base_page import TeachBasePage


class TeachDashboardPage(TeachBasePage):
    def __init__(self, browser):
        super(TeachDashboardPage, self).__init__(browser)

        assert self.on_correct_page("teach_dashboard_page")

    def go_to_class_page(self):
        self.browser.find_element(By.ID, "class_button").click()
        return TeachClassPage(self.browser)

    def open_school_tab(self):
        self.browser.find_element(By.ID, "tab-school").click()
        return self

    def open_classes_tab(self):
        self.browser.find_element(By.ID, "tab-classes").click()
        return self

    def open_account_tab(self):
        self.browser.find_element(By.ID, "tab-account").click()
        return self

    def check_organisation_details(self, details):
        correct = True

        first_field = list(details.items())[0][0]
        self.wait_for_element_by_id("id_" + first_field)

        for field, value in list(details.items()):
            correct &= self.browser.find_element(By.ID, "id_" + field).get_attribute("value") == value

        return correct

    def change_organisation_details(self, details):
        for field, value in list(details.items()):
            self.browser.find_element(By.ID, "id_" + field).clear()
            self.browser.find_element(By.ID, "id_" + field).send_keys(value)

        self.browser.find_element(By.ID, "update_details_button").click()
        return self

    def create_class(self, name, classmate_progress, teacher_id=None):
        self.browser.find_element(By.ID, "id_class_name").send_keys(name)
        if classmate_progress:
            self.browser.find_element(By.ID, "id_classmate_progress").click()

        if teacher_id is not None:
            Select(self.browser.find_element(By.ID, "id_teacher")).select_by_value(str(teacher_id))

        self.browser.find_element(By.ID, "create_class_button").click()

        return self

    def change_teacher_details(self, details):
        self._change_details(details)

        return self

    def change_email(self, first_name, last_name, new_email, password):
        self._change_details(
            {"first_name": first_name, "last_name": last_name, "email": new_email, "current_password": password}
        )

        return EmailVerificationNeededPage(self.browser)

    def change_password(self, first_name, last_name, new_password, password):
        self._change_details(
            {
                "first_name": first_name,
                "last_name": last_name,
                "password": new_password,
                "confirm_password": new_password,
                "current_password": password,
            }
        )

        from portal.tests.pageObjects.portal.teacher_login_page import \
            TeacherLoginPage

        return TeacherLoginPage(self.browser)

    def _change_details(self, details):
        for field, value in list(details.items()):
            self.browser.find_element(By.ID, "id_" + field).clear()
            self.browser.find_element(By.ID, "id_" + field).send_keys(value)
        self.browser.find_element(By.ID, "update_button").click()

    def check_account_details(self, details):
        correct = True

        for field, value in list(details.items()):
            correct &= self.browser.find_element(By.ID, "id_" + field).get_attribute("value") == value

        return correct

    def accept_independent_join_request(self):
        self.browser.find_element(By.ID, "allow_independent_button").click()
        return AddIndependentStudentToClassPage(self.browser)

    def deny_independent_join_request(self):
        self.browser.find_element(By.ID, "deny_independent_button").click()
        return self

    def has_independent_join_request(self, email):
        return self.element_exists_by_id("independent_request_table") and (
            email in self.browser.find_element(By.ID, "independent_request_table").text
        )

    def has_no_independent_join_requests(self):
        return self.element_does_not_exist_by_id("independent_request_table")

    def has_onboarding_complete_popup(self):
        return self.element_exists_by_id("info-popup") and (
            self.browser.find_element(By.ID, "info-popup").text.startswith("Registration complete!")
        )

    def is_teacher_in_school(self, name):
        return name in self.browser.find_element(By.ID, "teachers_table").text

    def is_not_teacher_in_school(self, name):
        return name not in self.browser.find_element(By.ID, "teachers_table").text

    def is_teacher_admin(self):
        return "Revoke admin" in self.browser.find_element(By.ID, "teachers_table").text

    def has_classes(self):
        return self.element_exists_by_id("classes-table")

    def does_not_have_classes(self):
        return self.element_does_not_exist_by_id("classes-table")

    def does_class_exist(self, name, access_code):
        return (
            self.has_classes()
            and (name in self.browser.find_element(By.ID, "classes-table").text)
            and (access_code in self.browser.find_element(By.ID, "classes-table").text)
        )

    def is_teacher_non_admin(self):
        return "Make admin" in self.browser.find_element(By.ID, "teachers_table").text

    def click_kick_button(self):
        self.browser.find_element(By.ID, "kick_button").click()

        return self

    def click_make_admin_button(self):
        self.browser.find_element(By.ID, "make_admin_button").click()

        return self

    def click_make_non_admin_button(self):
        self.browser.find_element(By.ID, "make_non_admin_button").click()

        return self

    def leave_organisation_with_students(self):
        self._click_leave_button()

        return TeachMoveClassesPage(self.browser)

    def _click_leave_button(self):
        self.browser.find_element(By.ID, "leave_organisation_button").click()

    def confirm_kick_with_students_dialog(self):
        self.confirm_dialog()

        return TeachMoveClassesPage(self.browser)



# Generated by Django 2.0 on 2020-10-28 17:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("portal", "0059_move_email_verifications_to_common"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="guardian",
            name="children",
        ),
        migrations.RemoveField(
            model_name="guardian",
            name="new_user",
        ),
        migrations.RemoveField(
            model_name="guardian",
            name="user",
        ),
        migrations.DeleteModel(
            name="Guardian",
        ),
    ]



import math

from common.helpers.emails import (
    send_verification_email,
)
from common.mail import campaign_ids, send_dotdigital_email
from common.models import DynamicElement, Student, Teacher, TotalActivity
from common.permissions import logged_in_as_student, logged_in_as_teacher
from common.utils import _using_two_factor
from django.contrib import messages as messages
from django.contrib.auth import logout
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.html import format_html
from django.views.decorators.cache import cache_control

from deploy import captcha
from portal.forms.play import IndependentStudentSignupForm
from portal.forms.teach import TeacherSignupForm
from portal.helpers.captcha import remove_captcha_from_forms
from portal.helpers.ratelimit import (
    RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_GROUP,
    RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_RATE,
    is_ratelimited,
)
from portal.strings.coding_club import CODING_CLUB_BANNER
from portal.strings.home_learning import HOME_LEARNING_BANNER
from portal.strings.ten_year_map import (
    TEN_YEAR_MAP_BANNER,
    TEN_YEAR_MAP_HEADLINE,
)


def register_view(request):
    if request.user.is_authenticated:
        return redirect_user_to_dashboard(request)
    else:
        return render_signup_form(request)


def redirect_user_to_dashboard(request):
    if logged_in_as_student(request.user):
        return HttpResponseRedirect(reverse_lazy("student_details"))
    elif logged_in_as_teacher(request.user):
        return HttpResponseRedirect(reverse_lazy("dashboard"))
    else:
        return HttpResponseRedirect(reverse_lazy("home"))


def logout_view(request):
    logout(request)
    return HttpResponseRedirect(reverse_lazy("home"))


def render_signup_form(request):
    invalid_form = False

    teacher_signup_form = TeacherSignupForm(prefix="teacher_signup")
    independent_student_signup_form = IndependentStudentSignupForm(
        prefix="independent_student_signup"
    )

    if request.method == "POST":
        if "teacher_signup-teacher_email" in request.POST:
            teacher_signup_form = TeacherSignupForm(
                request.POST, prefix="teacher_signup"
            )

            if not captcha.CAPTCHA_ENABLED:
                remove_captcha_from_forms(teacher_signup_form)

            if teacher_signup_form.is_valid():
                data = teacher_signup_form.cleaned_data
                return process_signup_form(request, data)

        else:
            independent_student_signup_form = IndependentStudentSignupForm(
                request.POST, prefix="independent_student_signup"
            )

            if independent_student_signup_form.is_valid():
                data = independent_student_signup_form.cleaned_data
                return process_independent_student_signup_form(request, data)

    res = render(
        request,
        "portal/register.html",
        {
            "teacher_signup_form": teacher_signup_form,
            "independent_student_signup_form": independent_student_signup_form,
        },
    )

    res.count = invalid_form
    return res


def _newsletter_ticked(form_data):
    return form_data["newsletter_ticked"]


def process_signup_form(request, data):
    email = data["teacher_email"]

    if email and User.objects.filter(email=email).exists():

        is_email_ratelimited = is_ratelimited(
            request=request,
            group=RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_GROUP,
            key=lambda *_: email,
            rate=RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_RATE,
            increment=True,
        )

        if not is_email_ratelimited:
            send_dotdigital_email(
                campaign_ids["user_already_registered"],
                [email],
                personalization_values={
                    "EMAIL": email,
                    "LOGIN_URL": request.build_absolute_uri(
                        reverse("teacher_login")
                    ),
                },
            )
    else:
        teacher = Teacher.objects.factory(
            first_name=data["teacher_first_name"],
            last_name=data["teacher_last_name"],
            email=data["teacher_email"],
            password=data["teacher_password"],
        )

        send_verification_email(request, teacher.user.user, data)

        TotalActivity.objects.update(
            teacher_registrations=F("teacher_registrations") + 1
        )

    return render(
        request,
        "portal/email_verification_needed.html",
        {"usertype": "TEACHER"},
        status=302,
    )


def process_independent_student_signup_form(request, data):
    email = data["email"]

    if email and User.objects.filter(email=email).exists():
        is_email_ratelimited = is_ratelimited(
            request=request,
            group=RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_GROUP,
            key=lambda *_: email,
            rate=RATELIMIT_USER_ALREADY_REGISTERED_EMAIL_RATE,
            increment=True,
        )

        if not is_email_ratelimited:
            send_dotdigital_email(
                campaign_ids["user_already_registered"],
                [email],
                personalization_values={
                    "EMAIL": email,
                    "LOGIN_URL": request.build_absolute_uri(
                        reverse("independent_student_login")
                    ),
                },
            )

        return render(
            request,
            "portal/email_verification_needed.html",
            {"usertype": "INDEP_STUDENT"},
            status=302,
        )

    student = Student.objects.independentStudentFactory(
        name=data["name"], email=data["email"], password=data["password"]
    )

    dob = data["date_of_birth"]
    age_in_days = timezone.now().date() - dob
    age = math.floor(age_in_days.days / 365.25)

    send_verification_email(request, student.new_user, data, age=age)

    TotalActivity.objects.update(
        independent_registrations=F("independent_registrations") + 1
    )

    return render(
        request,
        "portal/email_verification_needed.html",
        {"usertype": "INDEP_STUDENT"},
        status=302,
    )


def is_developer(request):
    return (
        hasattr(request.user, "userprofile")
        and request.user.userprofile.developer
    )


def redirect_teacher_to_correct_page(request, teacher):
    if teacher.has_school():
        link = reverse("two_factor:profile")
        if not _using_two_factor(request.user):
            messages.info(
                request,
                (
                    "You are not currently set up with two-factor authentication. "
                    + "Use your phone or tablet to enhance your account’s security.</br>"
                    + "Click <a href='"
                    + link
                    + "'>here</a> to find out more and "
                    + "set it up or go to your account page at any time."
                ),
                extra_tags="safe",
            )
        return reverse_lazy("dashboard")
    else:
        return reverse_lazy("onboarding-organisation")


@cache_control(private=True)
def home(request):
    # Putting this in a try catch because it causes some weird issue in the
    # tests where the first Selenium test passes, but any following test
    # fails because it cannot find the Maintenance banner instance.
    try:
        maintenance_banner = DynamicElement.objects.get(
            name="Maintenance banner"
        )

        if maintenance_banner.active:
            messages.info(
                request, format_html(maintenance_banner.text), extra_tags="safe"
            )
    except ObjectDoesNotExist:
        pass

    """
    This view is where we can add any messages to be shown upon loading the home page.
    Following this format:

    messages.success(request, "message text here", extra_tags="tag classes here")

    This example uses the success function which will display a welcoming message on the
    sub banner (right under the page header). Other functions can be used to indicate a
    warning, an error or a simple information.
    """
    return render(request, "portal/home.html")


def coding_club(request):
    return render(
        request, "portal/coding_club.html", {"BANNER": CODING_CLUB_BANNER}
    )


def home_learning(request):
    return render(
        request,
        "portal/home_learning.html",
        {"HOME_LEARNING_BANNER": HOME_LEARNING_BANNER},
    )


def ten_year_map_page(request):
    messages.info(
        request, "This page is currently under construction.", extra_tags="safe"
    )
    return render(
        request,
        "portal/ten_year_map.html",
        {"BANNER": TEN_YEAR_MAP_BANNER, "HEADLINE": TEN_YEAR_MAP_HEADLINE},
    )


def reset_screentime_warning(request):
    if request.user.is_authenticated:
        request.session["last_screentime_warning"] = timezone.now().timestamp()
    return HttpResponse(status=204)  # No content



from django import forms


class NewsletterForm(forms.Form):
    email = forms.EmailField(
        label="Sign up to receive updates about Code for Life games and teaching resources.",
        label_suffix="",
        widget=forms.EmailInput(
            attrs={
                "placeholder": "Your email address",
                "id": "newsletter_email_field",
            }
        ),
        help_text="Enter email address above",
    )

    age_verification = forms.BooleanField(
        widget=forms.CheckboxInput(), initial=False, required=True
    )


class ConsentForm(forms.Form):
    email = forms.EmailField(
        label="Email",
        label_suffix="",
        widget=forms.EmailInput(
            attrs={"placeholder": "your.name@yourdomain.com"}
        ),
    )

    consent_ticked = forms.BooleanField(
        widget=forms.CheckboxInput(), initial=False, required=True
    )



import ast
import re
from datetime import datetime

from common.helpers.emails import (
    NOTIFICATION_EMAIL,
    PASSWORD_RESET_EMAIL,
    delete_contact,
    send_email,
)
from common.mail import campaign_ids, send_dotdigital_email
from common.models import DailyActivity, Student, Teacher
from common.permissions import not_fully_logged_in, not_logged_in
from django.contrib import messages as messages
from django.contrib.auth import get_user_model
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.forms import PasswordResetForm, SetPasswordForm
from django.contrib.auth.models import User
from django.contrib.auth.tokens import default_token_generator
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_decode
from django.utils.translation import gettext as _
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters
from django.views.decorators.http import require_POST

from deploy import captcha
from portal import app_settings
from portal.forms.registration import (
    StudentPasswordResetForm,
    StudentPasswordResetSetPasswordForm,
    TeacherPasswordResetForm,
    TeacherPasswordResetSetPasswordForm,
)
from portal.helpers.captcha import remove_captcha_from_form
from portal.helpers.ratelimit import clear_ratelimit_cache_for_user
from portal.views.api import anonymise
from portal.views.login import has_user_lockout_expired


@user_passes_test(not_logged_in, login_url=reverse_lazy("home"))
def student_password_reset(request):
    usertype = "INDEP_STUDENT"
    return password_reset(
        request,
        usertype,
        from_email=PASSWORD_RESET_EMAIL,
        template_name="portal/reset_password.html",
        password_reset_form=StudentPasswordResetForm,
    )


def blocked_and_not_expired(user: Student or Teacher):
    return user.blocked_time and not has_user_lockout_expired(user)


def school_student_reset_password_tracker(request, activity_today):
    if "transfer_students" in request.POST:
        student_list = ast.literal_eval(request.POST.get("transfer_students", []))
        for student_id in student_list:
            current_student = Student.objects.get(id=student_id)
            if blocked_and_not_expired(current_student):
                activity_today.school_student_lockout_resets += 1
    elif "set_password" in request.POST:
        student_id = re.search("/(\d+)/", request.path).group(1)
        current_student = Student.objects.get(id=student_id)
        if blocked_and_not_expired(current_student):
            activity_today.school_student_lockout_resets += 1
    activity_today.save()


def teacher_or_indy_reset_password_tracker(request, activity_today, email):
    get_user = Teacher.objects.filter(new_user__email=email) or Student.objects.filter(new_user__email=email)
    if get_user.exists():
        user = get_user[0]
        if blocked_and_not_expired(user):
            if "teacher" in request.path:
                activity_today.teacher_lockout_resets += 1
            elif "student" in request.path:
                activity_today.indy_lockout_resets += 1
            activity_today.save()


def handle_reset_password_tracking(request, user_type, access_code=None, student_id=None):
    activity_today = DailyActivity.objects.get_or_create(date=datetime.now().date())[0]
    # school student has 2 different ways of resetting password
    # hence the function is extended
    # check for indy student or teacher account
    if user_type == "SCHOOL_STUDENT":
        school_student_reset_password_tracker(request, activity_today)
    elif "email" in request.POST:
        teacher_or_indy_reset_password_tracker(request, activity_today, request.POST.get("email", ""))


@user_passes_test(not_fully_logged_in, login_url=reverse_lazy("teacher_login"))
def teacher_password_reset(request):
    usertype = "TEACHER"
    return password_reset(
        request,
        usertype,
        from_email=PASSWORD_RESET_EMAIL,
        template_name="portal/reset_password.html",
        password_reset_form=TeacherPasswordResetForm,
    )


@csrf_protect
def password_reset(
    request,
    usertype,
    template_name="portal/reset_password.html",
    email_template_name="portal/reset_password_email.html",
    subject_template_name="registration/password_reset_subject.txt",
    password_reset_form=PasswordResetForm,
    token_generator=default_token_generator,
    from_email=None,
    current_app=None,
    extra_context=None,
    html_email_template_name=None,
):
    if request.method == "POST":
        handle_reset_password_tracking(request, usertype)
        form = password_reset_form(request.POST)
        if not captcha.CAPTCHA_ENABLED:
            remove_captcha_from_form(form)
        if form.is_valid():
            opts = {
                "use_https": request.is_secure(),
                "token_generator": token_generator,
                "from_email": from_email,
                "email_template_name": email_template_name,
                "subject_template_name": subject_template_name,
                "request": request,
                "html_email_template_name": html_email_template_name,
            }
            form.save(**opts)

            return render(request, "portal/reset_password_email_sent.html", {"usertype": usertype})
    else:
        form = password_reset_form()

    if not captcha.CAPTCHA_ENABLED:
        remove_captcha_from_form(form)

    context = {
        "form": form,
        "title": _("Password reset"),
        "settings": app_settings,
        "should_use_recaptcha": captcha.CAPTCHA_ENABLED,
        "usertype": usertype,
    }

    update_context_and_apps(request, context, current_app, extra_context)

    return TemplateResponse(request, template_name, context)


def update_context_and_apps(request, context, current_app, extra_context):
    if extra_context is not None:
        context.update(extra_context)

    if current_app is not None:
        request.current_app = current_app


def password_reset_done(
    request,
    template_name="portal/reset_password_email_sent.html",
    current_app=None,
    extra_context=None,
):
    context = {"title": _("Password reset sent")}

    update_context_and_apps(request, context, current_app, extra_context)

    return TemplateResponse(request, template_name, context)


@sensitive_post_parameters()
@never_cache
def password_reset_confirm(
    request,
    usertype,
    uidb64=None,
    token=None,
    template_name="portal/reset_password_confirm.html",
    token_generator=default_token_generator,
    set_password_form=SetPasswordForm,
    current_app=None,
    extra_context=None,
):
    """
    View that checks the hash in a password reset link and presents a
    form for entering a new password.
    """
    UserModel = get_user_model()
    check_uidb64(uidb64, token)

    try:
        uid = force_bytes(urlsafe_base64_decode(uidb64))
        user = UserModel._default_manager.get(pk=uid)
    except (TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
        user = None

    if user_is_authenticated(user, token_generator, token):
        validlink = True
        title = _("Enter new password")
        if request.method == "POST":
            form = set_password_form(user, request.POST)
            if form.is_valid():
                form.save()

                # Reset ratelimit cache upon successful password reset
                clear_ratelimit_cache_for_user(user.username)

                _check_and_unblock_user(user.username, usertype)

                return render(request, "portal/reset_password_done.html", {"usertype": usertype})
        else:
            form = set_password_form(user)
    else:
        validlink = False
        form = None
        title = _("Password reset unsuccessful")

    context = {"form": form, "title": title, "validlink": validlink}

    update_context_and_apps(request, context, current_app, extra_context)

    return TemplateResponse(request, template_name, context)


def _check_and_unblock_user(username, usertype, access_code=None):
    if usertype == "TEACHER":
        user = Teacher.objects.get(new_user__username=username)
    elif usertype == "INDEP_STUDENT":
        user = Student.objects.get(new_user__username=username)
    elif usertype == "SCHOOL_STUDENT":
        user = Student.objects.get(new_user__first_name=username, class_field__access_code=access_code)

    if user.blocked_time is not None:
        user.blocked_time = None
        user.save()


def check_uidb64(uidb64, token):
    assert uidb64 is not None and token is not None  # checked by URLconf


def user_is_authenticated(user, token_generator, token):
    return user is not None and token_generator.check_token(user, token)


@user_passes_test(not_fully_logged_in, login_url=reverse_lazy("home"))
def password_reset_check_and_confirm(request, uidb64=None, token=None):
    """
    Customised standard django auth view with customised form to incorporate checking the password set is strong enough
    """
    UserModel = get_user_model()
    try:
        uid = urlsafe_base64_decode(uidb64)
        user = UserModel._default_manager.get(pk=uid)
    except (TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
        user = None
    if user and hasattr(user, "new_student"):
        usertype = "INDEP_STUDENT"
        return password_reset_confirm(
            request,
            usertype,
            set_password_form=StudentPasswordResetSetPasswordForm,
            uidb64=uidb64,
            token=token,
            extra_context={"usertype": usertype},
        )
    else:
        usertype = "TEACHER"
        return password_reset_confirm(
            request,
            usertype,
            set_password_form=TeacherPasswordResetSetPasswordForm,
            uidb64=uidb64,
            token=token,
            extra_context={"usertype": usertype},
        )


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
def delete_account(request):
    user = request.user
    password = request.POST.get("password")

    if not user.check_password(password):
        messages.error(request, "Your account was not deleted due to incorrect password.")
        return HttpResponseRedirect(reverse_lazy("dashboard"))

    email = user.email
    anonymise(user)

    # remove from dotmailer
    if bool(request.POST.get("unsubscribe_newsletter")):
        delete_contact(email)

    # send confirmation email
    send_dotdigital_email(campaign_ids["delete_account"], [email])

    return HttpResponseRedirect(reverse_lazy("home"))



import pytest
from common.utils import field_exists


@pytest.mark.django_db
def test_preview_user_field_removed(migrator):
    migrator.apply_initial_migration(("portal", "0055_add_preview_user"))
    new_state = migrator.apply_tested_migration(("portal", "0056_remove_preview_user"))

    userprofile_model = new_state.apps.get_model("portal", "UserProfile")
    assert not field_exists(userprofile_model, "preview_user")

    school_model = new_state.apps.get_model("portal", "School")
    assert not field_exists(school_model, "eligible_for_testing")



import os

from django.conf import settings
from django.http import HttpResponse


class BasicAuthMiddleware:
    def unauthed(self):
        response = HttpResponse(
            """<html><title>Auth required</title><body>
                                <h1>Authorization Required</h1></body></html>""",
            content_type="text/html",
        )
        response["WWW-Authenticate"] = (
            'Basic realm="' + os.getenv("APPLICATION_ID", "") + '"'
        )
        response.status_code = 401
        return response

    def process_request(self, request):
        if request.META.has_key("HTTP_AUTHORIZATION"):
            authmeth, auth = request.META["HTTP_AUTHORIZATION"].split(" ", 1)
            if authmeth.lower() == "basic":
                auth = auth.strip().decode("base64")
                username, password = auth.split(":", 1)
                if (
                    username != settings.BASICAUTH_USERNAME
                    or password != settings.BASICAUTH_PASSWORD
                ):
                    return self.unauthed()
            else:
                return self.unauthed()
        else:
            return self.unauthed()



from __future__ import absolute_import

from selenium.webdriver.common.by import By

from . import class_page, dashboard_page
from .teach_base_page import TeachBasePage


class OnboardingStudentListPage(TeachBasePage):
    def __init__(self, browser):
        super(OnboardingStudentListPage, self).__init__(browser)

        assert self.on_correct_page("onboarding_student_list_page")

    def student_exists(self, name):
        return name in self.browser.find_element(By.ID, "student_table").text

    def is_student_password(self, password):
        return password in self.browser.find_element(By.XPATH, "//table[@id='student_table']/tbody/tr[4]/td[2]").text

    def go_back_to_class(self):
        self.browser.find_element(By.ID, "back_to_class_button").click()
        return class_page.TeachClassPage(self.browser)

    def get_first_login_url(self):
        return self.browser.find_element(By.XPATH, "//table[@id='student_table']/tbody/tr[4]/td[4]/div/div[1]").text

    def complete_setup(self):
        self.browser.find_element(By.ID, "complete_setup_button").click()
        return dashboard_page.TeachDashboardPage(self.browser)



CODING_CLUB_BANNER = {
    "title": "Coding clubs",
    "subtitle": "A FREE set of slides and guides to run your own coding clubs",
    "text": "",
    "button": "",
    "image_class": "banner--picture--coding-club",
    "alt": "Boy using a laptop",
    "image_description": "Credit: Jeswin Thomas, Unsplash",
}



import pytest


@pytest.mark.django_db
def test_portaladmin_has_teacher_profile(migrator):
    migrator.apply_initial_migration(("portal", "0060_delete_guardian"))
    new_state = migrator.apply_tested_migration(("portal", "0061_make_portaladmin_teacher"))

    User = new_state.apps.get_model("auth", "User")
    UserProfile = new_state.apps.get_model("common", "UserProfile")
    School = new_state.apps.get_model("common", "School")
    Teacher = new_state.apps.get_model("common", "Teacher")
    Class = new_state.apps.get_model("common", "Class")
    Student = new_state.apps.get_model("common", "Student")

    portaladmin = User.objects.get(username="portaladmin")

    assert portaladmin.first_name == "Portal"
    assert portaladmin.last_name == "Admin"
    assert portaladmin.email == "codeforlife-portal@ocado.com"

    portaladmin_userprofile = UserProfile.objects.get(user=portaladmin)
    portaladmin_school = School.objects.get(name="Swiss Federal Polytechnic")

    portaladmin_teacher = Teacher.objects.get(new_user=portaladmin)

    assert portaladmin_teacher.user == portaladmin_userprofile
    assert portaladmin_teacher.school == portaladmin_school

    portaladmin_class = Class.objects.get(access_code="PO123")

    assert portaladmin_class.teacher == portaladmin_teacher

    portaladmin_student_user = User.objects.get(username="portaladmin student")
    portaladmin_student_userprofile = UserProfile.objects.get(user=portaladmin_student_user)
    portaladmin_student = Student.objects.get(user=portaladmin_student_userprofile)

    assert portaladmin_student.class_field == portaladmin_class


@pytest.mark.django_db
def test_make_portaladmin_teacher_rollback(migrator):
    migrator.apply_initial_migration(("portal", "0061_make_portaladmin_teacher"))
    new_state = migrator.apply_tested_migration(("portal", "0060_delete_guardian"))

    User = new_state.apps.get_model("auth", "User")
    UserProfile = new_state.apps.get_model("common", "UserProfile")
    Teacher = new_state.apps.get_model("common", "Teacher")
    Class = new_state.apps.get_model("common", "Class")

    portaladmin = User.objects.get(username="portaladmin")

    assert portaladmin.first_name == ""
    assert portaladmin.last_name == ""
    assert portaladmin.email == "('codeforlife-portal@ocado.com',)"

    assert not UserProfile.objects.filter(user=portaladmin).exists()
    assert not Teacher.objects.filter(new_user=portaladmin).exists()
    assert not Class.objects.filter(access_code="PO123").exists()
    assert not User.objects.filter(username="portaladmin student").exists()



from django import template

register = template.Library()


@register.inclusion_tag("portal/partials/banner.html", takes_context=True)
def banner(context, banner_name):
    """
    Registers the inclusion tag for the banner partial.
    Takes in the name of the banner.
    The template currently expects the following context elements:
    - title: the heading of the banner
    - subtitle (optional): a smaller heading below the title
    - text (optional): a description paragraph below the subtitle
    - button (optional): a dictionary containing the text and link of a button below
    the text elements
    - image_class: the CSS class of the image to be shown in the hexagon
    - image_description (optional): an image description that appears when hovering over the image
    """
    return context[banner_name]



from common.helpers.emails import (
    add_to_dotmailer,
    get_dotmailer_user_by_email,
    send_dotmailer_consent_confirmation_email_to_user,
    add_consent_record_to_dotmailer_user,
    DotmailerUserType,
)
from common.mail import address_book_ids
from django.contrib import messages as messages
from django.http import HttpResponseRedirect, HttpResponse
from django.shortcuts import render
from django.urls import reverse_lazy
from django.views.decorators.csrf import csrf_exempt

from portal.forms.dotmailer import NewsletterForm, ConsentForm


@csrf_exempt
def process_newsletter_form(request):
    if request.method == "POST":
        newsletter_form = NewsletterForm(data=request.POST)
        if newsletter_form.is_valid():
            user_email = newsletter_form.cleaned_data["email"]
            add_to_dotmailer(
                "",
                "",
                user_email,
                address_book_ids["newsletter"],
                DotmailerUserType.NO_ACCOUNT,
            )
            messages.success(request, "Thank you for signing up! 🎉")
            return HttpResponseRedirect(reverse_lazy("home"))
        messages.error(
            request,
            "Invalid email address. Please try again.",
            extra_tags="sub-nav--warning",
        )
        return HttpResponseRedirect(reverse_lazy("home"))

    return HttpResponse(status=405)


def dotmailer_consent_form(request):
    if request.method == "POST":
        consent_form = ConsentForm(data=request.POST)

        if consent_form.is_valid():
            user_email = consent_form.cleaned_data["email"]
            user = get_dotmailer_user_by_email(user_email)
            try:
                add_consent_record_to_dotmailer_user(user)
            except KeyError:
                # if no user is registered with that email, show error message
                pass
            else:
                # no error
                send_dotmailer_consent_confirmation_email_to_user(user)
                return HttpResponseRedirect(reverse_lazy("home"))

        messages.error(
            request,
            "Valid email address and consent required. Please try again.",
            extra_tags="sub-nav--warning",
        )
        return HttpResponseRedirect(reverse_lazy("consent_form"))

    else:
        consent_form = ConsentForm()

    return render(
        request,
        "portal/dotmailer_consent_form.html",
        {"form": consent_form},
    )



from datetime import datetime, timedelta

import pytz
from common.models import Teacher, Student
from django.shortcuts import redirect
from django.urls import reverse_lazy


def old_login_form_redirect(request):
    return redirect(reverse_lazy("home"))


def has_user_lockout_expired(user: Teacher or Student) -> bool:
    return datetime.now(tz=pytz.utc) - user.blocked_time > timedelta(hours=24)



import re
from datetime import timedelta, date

from common.helpers.emails import send_verification_email
from common.models import Class, Student, stripStudentName
from common.permissions import logged_in_as_independent_student
from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth.forms import AuthenticationForm
from django.utils import timezone
from django_recaptcha.fields import ReCaptchaField
from django_recaptcha.widgets import ReCaptchaV2Invisible

from portal.forms.error_messages import INVALID_LOGIN_MESSAGE
from portal.helpers.password import PasswordStrength, form_clean_password
from portal.helpers.regexes import ACCESS_CODE_PATTERN


class StudentClassCodeForm(forms.Form):
    access_code = forms.CharField(
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Class code"}),
        help_text="Enter your class code",
    )

    def clean(self):
        access_code = self.cleaned_data.get("access_code", None)

        if access_code:
            if re.fullmatch(ACCESS_CODE_PATTERN, access_code.upper()) is None:
                raise forms.ValidationError("Uh oh! You didn't input a valid class code.")

        return self.cleaned_data


class StudentLoginForm(AuthenticationForm):
    username = forms.CharField(
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Username"}),
        help_text="Enter your username",
    )
    password = forms.CharField(
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
        help_text="Enter your password",
    )

    def __init__(self, *args, **kwargs):
        self.access_code = kwargs.pop("access_code", None)
        super(StudentLoginForm, self).__init__(*args, **kwargs)

    def clean(self):
        name = self.cleaned_data.get("username", None)
        password = self.cleaned_data.get("password", None)

        if name and self.access_code and password:
            student, user = self.check_for_errors(name, self.access_code, password)

            self.student = student
            self.user_cache = user
        return self.cleaned_data

    def check_for_errors(self, name, access_code, password):
        classes = Class.objects.filter(access_code__iexact=access_code)
        if len(classes) != 1:
            raise forms.ValidationError("Invalid name, class access code or password")
        klass = classes[0]

        name = stripStudentName(name)

        students = Student.objects.filter(new_user__first_name__iexact=name, class_field=klass)
        if len(students) != 1:
            raise forms.ValidationError("Invalid name, class access code or password")

        student = students[0]
        user = authenticate(username=student.new_user.username, password=password.lower())

        # Try the case sensitive password too, for previous accounts that don't have the lowercase one stored
        if user is None:
            user = authenticate(username=student.new_user.username, password=password)

        if user is None:
            raise forms.ValidationError("Invalid name, class access code or password")
        if not user.is_active:
            raise forms.ValidationError("This user account has been deactivated")

        return student, user


class StudentEditAccountForm(forms.Form):
    password = forms.CharField(
        required=True,
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "New password"}),
        help_text="Enter new password",
    )
    confirm_password = forms.CharField(
        required=True,
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Confirm new password"}),
        help_text="Confirm new password",
    )
    current_password = forms.CharField(
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Current password"}),
        help_text="Enter your current password",
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(StudentEditAccountForm, self).__init__(*args, **kwargs)

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.STUDENT)

    def clean(self):
        return clean_confirm_password(self, independent=False)


class IndependentStudentEditAccountForm(forms.Form):
    name = forms.CharField(
        max_length=100,
        required=False,
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Name"}),
        help_text="Enter your name",
    )
    email = forms.EmailField(
        required=False,
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "New email address (optional)"}),
        help_text="Enter new email address (optional)",
    )
    password = forms.CharField(
        required=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "New password (optional)"}),
        help_text="Enter new password (optional)",
    )
    confirm_password = forms.CharField(
        label="Confirm new password",
        required=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Confirm new password"}),
        help_text="Confirm new password",
    )
    current_password = forms.CharField(
        label="Current password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Current password"}),
        help_text="Enter your current password",
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(IndependentStudentEditAccountForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        name = self.cleaned_data.get("name", None)
        if not self.user.new_student.class_field:
            if name == "":
                raise forms.ValidationError("This field is required")

            if re.match(re.compile("^[\w ]+$"), name) is None:
                raise forms.ValidationError("Names may only contain letters, numbers, dashes, underscores, and spaces.")

        return name

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.INDEPENDENT)

    def clean(self):
        return clean_confirm_password(self, independent=True)


def clean_confirm_password(self, independent=True):
    password = self.cleaned_data.get("password", None)
    confirm_password = self.cleaned_data.get("confirm_password", None)
    current_password = self.cleaned_data.get("current_password", None)

    # Password is lowercase for non-independent students
    if not independent:
        if password is not None:
            password = password.lower()
        if confirm_password is not None:
            confirm_password = confirm_password.lower()

    if are_password_and_confirm_password_different(password, confirm_password):
        raise forms.ValidationError("Your new passwords do not match")

    if current_password and not self.user.check_password(current_password):
        # If it's not an independent student, check their lowercase password as well
        if independent or not self.user.check_password(current_password.lower()):
            raise forms.ValidationError("Your current password was incorrect")

    return self.cleaned_data


def are_password_and_confirm_password_different(password, confirm_password):
    return password is not None and password != confirm_password


class IndependentStudentSignupForm(forms.Form):
    date_of_birth = forms.DateField(
        help_text="Please enter your date of birth (we do not store this information).",
        widget=forms.SelectDateWidget(
            years=range(date.today().year, date.today().year - 100, -1), empty_label=("Year", "Month", "Day")
        ),
        required=False,
    )

    name = forms.CharField(
        max_length=100,
        help_text="Enter full name",
        widget=forms.TextInput(attrs={"autocomplete": "off", "placeholder": "Full name"}),
    )

    email = forms.EmailField(
        help_text="Enter your email address",
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
    )

    consent_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=True)
    newsletter_ticked = forms.BooleanField(widget=forms.CheckboxInput(), initial=False, required=False)

    password = forms.CharField(
        help_text="Enter a password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
    )

    confirm_password = forms.CharField(
        help_text="Repeat password",
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Repeat password"}),
    )

    captcha = ReCaptchaField(widget=ReCaptchaV2Invisible)

    def clean_name(self):
        name = self.cleaned_data.get("name", None)
        if re.match(re.compile("^[\w ]+$"), name) is None:
            raise forms.ValidationError("Names may only contain letters, numbers, dashes, underscores, and spaces.")

        return name

    def clean_password(self):
        return form_clean_password(self, "password", PasswordStrength.INDEPENDENT)

    def clean(self):
        password = self.cleaned_data.get("password", None)
        confirm_password = self.cleaned_data.get("confirm_password", None)

        if password and confirm_password and password != confirm_password:
            raise forms.ValidationError("Your passwords do not match")

        return self.cleaned_data


class IndependentStudentLoginForm(AuthenticationForm):
    username = forms.EmailField(
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
        help_text="Enter your email address",
    )
    password = forms.CharField(
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Password"}),
        help_text="Enter your password",
    )

    def clean(self):
        super().clean()

    def confirm_login_allowed(self, user):
        if not logged_in_as_independent_student(user):
            self.show_invalid_login_message()

        if not user.userprofile.is_verified:
            send_verification_email(self.request, user, self.data)
            self.show_invalid_login_message()

    def get_invalid_login_error(self):
        self.show_invalid_login_message()

    def show_invalid_login_message(self):
        raise forms.ValidationError(INVALID_LOGIN_MESSAGE)


class StudentJoinOrganisationForm(forms.Form):
    access_code = forms.CharField(label="Class Access Code", widget=forms.TextInput(attrs={"placeholder": "AB123"}))

    def clean(self):
        access_code = self.cleaned_data.get("access_code", None)
        join_error_text = "The class code you entered either does not exist or is not currently accepting join requests. Please double check that you have entered the correct class code and contact the teacher of the class to ensure their class is currently accepting join requests."

        if access_code:
            classes = Class.objects.filter(access_code=access_code)
            if len(classes) != 1:
                raise forms.ValidationError(join_error_text)

            self.klass = classes[0]

            if not self.klass.always_accept_requests and (
                self.klass.accept_requests_until is None
                or self.klass.accept_requests_until - timezone.now()
                < timedelta()
            ):
                raise forms.ValidationError(join_error_text)
        return self.cleaned_data



from django.shortcuts import render


def privacy_notice(request):
    return render(
        request,
        "portal/privacy_notice.html",
        {"last_updated": "19th March 2025", "last_updated_children": "19th March 2025"},
    )


def terms(request):
    return render(request, "portal/terms.html", {"last_updated": "11th July 2022"})



from common.models import Student, UserSession
from django.contrib import messages
from django.contrib.auth.models import User
from django.contrib.auth.views import LoginView
from django.shortcuts import render
from django.urls import reverse_lazy

from portal.forms.play import IndependentStudentLoginForm
from portal.helpers.ratelimit import clear_ratelimit_cache_for_user
from . import has_user_lockout_expired


class IndependentStudentLoginView(LoginView):
    template_name = "portal/login/independent_student.html"
    form_class = IndependentStudentLoginForm
    success_url = reverse_lazy("independent_student_details")
    join_class_url = reverse_lazy("student_join_organisation")
    extra_context = {"usertype": "INDEP_STUDENT"}

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or self.success_url

    def _add_logged_in_as_message(self, request):
        messages.info(
            request,
            f"<strong>You are logged in as an independent student. If you want to join "
            f"a school, you need to <a href='{self.join_class_url}' id='student_join_school_link'> "
            f"request to join one</a>.</strong>",
            extra_tags="safe",
        )

    def post(self, request, *args, **kwargs):
        """
        If the email address inputted in the form corresponds to that of a blocked
        account, this redirects the user to the locked out page. However, if the lockout
        time is more than 24 hours before this is executed, the account is unlocked.
        """
        email = request.POST.get("username")
        if Student.objects.filter(new_user__email=email).exists():
            student = Student.objects.get(new_user__email=email)

            if student.blocked_time is not None:
                if has_user_lockout_expired(student):
                    student.blocked_time = None
                    student.save()
                else:
                    return render(
                        self.request,
                        "portal/locked_out.html",
                        {"is_teacher": False},
                    )

        return super(IndependentStudentLoginView, self).post(request, *args, **kwargs)

    def form_valid(self, form):
        self._add_logged_in_as_message(self.request)

        # Reset ratelimit cache upon successful login
        clear_ratelimit_cache_for_user(form.cleaned_data["username"])

        # Log the login time
        email = self.request.POST.get("username")
        user = User.objects.get(email=email)
        session = UserSession(user=user)
        session.save()

        return super().form_valid(form)



import socket
import time

from django.contrib.sites.models import Site
from django.urls import reverse

from deploy import captcha
from portal.tests.pageObjects.portal.home_page import HomePage
from .selenium_test_case import SeleniumTestCase


class BaseTest(SeleniumTestCase):
    @classmethod
    def setUpClass(cls):
        cls.orig_captcha_enabled = captcha.CAPTCHA_ENABLED
        captcha.CAPTCHA_ENABLED = False
        super(BaseTest, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        captcha.CAPTCHA_ENABLED = cls.orig_captcha_enabled
        super(BaseTest, cls).tearDownClass()

    def go_to_homepage(self):
        path = reverse("home")
        self._go_to_path(path)
        return HomePage(self.selenium)

    def _go_to_path(self, path):
        socket.setdefaulttimeout(20)
        attempts = 0
        while attempts <= 3:
            try:
                self.selenium.get(self.live_server_url + path)
            except socket.timeout:
                if attempts > 2:
                    raise
                time.sleep(10)
            else:
                break
            attempts += 1

    def __call__(self, result=None):
        self._set_site_to_local_domain()
        return super().__call__(result)

    def _set_site_to_local_domain(self):
        """
        Sets the Site Django object to the local domain (locally, localhost:8000).
        Needed to generate valid registration and password reset links in tests.
        """
        current_site = Site.objects.get_current()
        current_site.domain = f"{self.server_thread.host}:{self.server_thread.port}"
        current_site.save()



from common.permissions import teacher_verified
from django.http import HttpResponse
from django.urls import include, path, re_path
from django.views.generic import RedirectView
from django.views.generic.base import TemplateView
from django.views.i18n import JavaScriptCatalog
from game.views.level import play_default_level
from two_factor.views import (
    BackupTokensView,
    ProfileView,
    QRGeneratorView,
    SetupCompleteView,
)

from portal.helpers.decorators import ratelimit
from portal.helpers.ratelimit import (
    RATELIMIT_LOGIN_GROUP,
    RATELIMIT_LOGIN_RATE,
    RATELIMIT_LOGIN_RATE_SCHOOL_STUDENT,
    RATELIMIT_METHOD,
    school_student_key,
)
from portal.helpers.regexes import ACCESS_CODE_REGEX, JWT_REGEX
from portal.views import cron, google_analytics
from portal.views.about import about, contribute, getinvolved
from portal.views.admin import AdminChangePasswordDoneView, AdminChangePasswordView
from portal.views.api import (
    AnonymiseOrphanSchoolsView,
    InactiveUsersView,
    RemoveFakeAccounts,
    last_connected_since,
    number_users_per_country,
    registered_users,
)
from portal.views.dotmailer import dotmailer_consent_form, process_newsletter_form
from portal.views.email import verify_email
from portal.views.home import (
    coding_club,
    home,
    home_learning,
    logout_view,
    register_view,
    reset_screentime_warning,
    ten_year_map_page,
)
from portal.views.legal import privacy_notice, terms
from portal.views.login import old_login_form_redirect
from portal.views.login.independent_student import IndependentStudentLoginView
from portal.views.login.student import (
    StudentClassCodeView,
    StudentLoginView,
    student_direct_login,
)
from portal.views.login.teacher import TeacherLoginView
from portal.views.organisation import organisation_leave, organisation_manage
from portal.views.play_landing_page import play_landing_page
from portal.views.registration import (
    delete_account,
    password_reset_check_and_confirm,
    password_reset_done,
    student_password_reset,
    teacher_password_reset,
)
from portal.views.student.edit_account_details import (
    SchoolStudentEditAccountView,
    independentStudentEditAccountView,
    student_edit_account,
)
from portal.views.student.play import (
    IndependentStudentDashboard,
    SchoolStudentDashboard,
    student_join_organisation,
)
from portal.views.teach import teach
from portal.views.teacher.dashboard import (
    dashboard_manage,
    delete_teacher_invite,
    invite_toggle_admin,
    invited_teacher,
    organisation_kick,
    organisation_toggle_admin,
    resend_invite_teacher,
    teacher_accept_student_request,
    teacher_disable_2FA,
    teacher_reject_student_request,
)
from portal.views.teacher.teach import (
    teacher_class_password_reset,
    teacher_delete_class,
    teacher_delete_students,
    teacher_dismiss_students,
    teacher_download_csv,
    teacher_edit_class,
    teacher_edit_student,
    teacher_move_students,
    teacher_move_students_to_class,
    teacher_onboarding_create_class,
    teacher_onboarding_edit_class,
    teacher_print_reminder_cards,
    teacher_view_class,
)
from portal.views.two_factor.core import CustomSetupView
from portal.views.two_factor.profile import CustomDisableView

js_info_dict = {"packages": ("conf.locale",)}

two_factor_patterns = [
    re_path(r"^account/two_factor/setup/$", CustomSetupView.as_view(), name="setup"),
    re_path(r"^account/two_factor/qrcode/$", QRGeneratorView.as_view(), name="qr"),
    re_path(
        r"^account/two_factor/setup/complete/$",
        SetupCompleteView.as_view(),
        name="setup_complete",
    ),
    re_path(
        r"^account/two_factor/backup/tokens/$",
        teacher_verified(BackupTokensView.as_view()),
        name="backup_tokens",
    ),
    re_path(
        r"^account/two_factor/$",
        teacher_verified(ProfileView.as_view()),
        name="profile",
    ),
    re_path(
        r"^account/two_factor/disable/$",
        teacher_verified(CustomDisableView.as_view()),
        name="disable",
    ),
]


urlpatterns = [
    path(
        "google-analytics/collect/",
        google_analytics.collect,
        name="collect-google-analytics",
    ),
    path(
        "cron/",
        include(
            [
                path(
                    "user/",
                    include(
                        [
                            path(
                                "unverified/send-first-reminder/",
                                cron.user.FirstVerifyEmailReminderView.as_view(),
                                name="first-verify-email-reminder",
                            ),
                            path(
                                "unverified/send-second-reminder/",
                                cron.user.SecondVerifyEmailReminderView.as_view(),
                                name="second-verify-email-reminder",
                            ),
                            path(
                                "unverified/delete/",
                                cron.user.AnonymiseUnverifiedAccounts.as_view(),
                                name="anonymise-unverified-accounts",
                            ),
                            path(
                                "inactive/send-first-reminder/",
                                cron.user.FirstInactivityReminderView.as_view(),
                                name="first-inactivity-reminder",
                            ),
                            path(
                                "inactive/send-second-reminder/",
                                cron.user.SecondInactivityReminderView.as_view(),
                                name="second-inactivity-reminder",
                            ),
                            path(
                                "inactive/send-final-reminder/",
                                cron.user.FinalInactivityReminderView.as_view(),
                                name="final-inactivity-reminder",
                            ),
                        ]
                    ),
                ),
            ]
        ),
    ),
    re_path(
        r"^favicon\.ico$",
        RedirectView.as_view(url="/static/portal/img/favicon.ico", permanent=True),
    ),
    re_path(
        r"^administration/password_change/$",
        AdminChangePasswordView.as_view(),
        name="administration_password_change",
    ),
    re_path(
        r"^administration/password_change_done/$",
        AdminChangePasswordDoneView.as_view(),
        name="administration_password_change_done",
    ),
    re_path(r"^users/inactive/", InactiveUsersView.as_view(), name="inactive_users"),
    re_path(
        r"^locked_out/$",
        TemplateView.as_view(template_name="portal/locked_out.html"),
        name="locked_out",
    ),
    re_path(
        r"^",
        include((two_factor_patterns, "two_factor"), namespace="two_factor"),
    ),
    re_path(r"^i18n/", include("django.conf.urls.i18n")),
    re_path(r"^jsi18n/$", JavaScriptCatalog.as_view(), js_info_dict),
    re_path(
        r"^(?P<level_name>[A-Z0-9]+)/$",
        play_default_level,
        name="play_default_level",
    ),
    re_path(r"^$", home, name="home"),
    re_path(r"^home-learning", home_learning, name="home-learning"),
    re_path(r"^register_form", register_view, name="register"),
    re_path(
        r"^login/teacher/$",
        # The ratelimit decorator checks how often a POST request is performed on that view.
        # It checks against the username value specifically. If the number of requests
        # exceeds the specified rate, then the user will be blocked (if block = True).
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key="post:auth-username",
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE,
            block=True,
        )(TeacherLoginView.as_view()),
        name="teacher_login",
    ),
    re_path(
        rf"^login/student/(?P<access_code>{ACCESS_CODE_REGEX})/(?:(?P<login_type>classform)/)?$",
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key=school_student_key,
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE_SCHOOL_STUDENT,
            block=True,
            is_teacher=False,
        )(StudentLoginView.as_view()),
        name="student_login",
    ),
    re_path(
        r"^login/student/$",
        StudentClassCodeView.as_view(),
        name="student_login_access_code",
    ),
    re_path(
        r"^u/(?P<user_id>[0-9]+)/(?P<login_id>[a-z0-9]+)/$",
        student_direct_login,
        name="student_direct_login",
    ),
    re_path(
        r"^login/independent/$",
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key="post:username",
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE,
            block=True,
            is_teacher=False,
        )(IndependentStudentLoginView.as_view()),
        name="independent_student_login",
    ),
    re_path(r"^login_form", old_login_form_redirect, name="old_login_form"),
    re_path(r"^logout/$", logout_view, name="logout_view"),
    re_path(
        r"^news_signup/$",
        process_newsletter_form,
        name="process_newsletter_form",
    ),
    re_path(r"^consent_form/$", dotmailer_consent_form, name="consent_form"),
    re_path(
        r"^verify_email/$",
        TemplateView.as_view(template_name="portal/email_verification_needed.html"),
        name="email_verification",
    ),
    re_path(
        rf"^verify_email/(?P<token>{JWT_REGEX})/$",
        verify_email,
        name="verify_email",
    ),
    re_path(
        r"^user/password/reset/student/$",
        student_password_reset,
        name="student_password_reset",
    ),
    re_path(
        r"^user/password/reset/teacher/$",
        teacher_password_reset,
        name="teacher_password_reset",
    ),
    re_path(
        r"^user/password/reset/done/$",
        password_reset_done,
        name="reset_password_email_sent",
    ),
    re_path(
        r"^user/password/reset/(?P<uidb64>[0-9A-Za-z]+)-(?P<token>.+)/$",
        password_reset_check_and_confirm,
        name="password_reset_check_and_confirm",
    ),
    re_path(
        r"^user/reset_screentime_warning/$",
        reset_screentime_warning,
        name="reset_screentime_warning",
    ),
    re_path(
        r"^user/reset_session_time/$",
        lambda _: HttpResponse(status=204),
        name="reset_session_time",
    ),
    re_path(
        r"^teacher/password/reset/complete/$",
        TemplateView.as_view(template_name="portal/reset_password_done.html"),
        name="password_reset_complete",
    ),
    re_path(r"^teach/$", teach, name="teach"),
    re_path(
        r"^teach/onboarding-organisation/$",
        organisation_manage,
        name="onboarding-organisation",
    ),
    re_path(
        r"^teach/onboarding-classes",
        teacher_onboarding_create_class,
        name="onboarding-classes",
    ),
    re_path(
        rf"^teach/onboarding-class/(?P<access_code>{ACCESS_CODE_REGEX})$",
        teacher_onboarding_edit_class,
        name="onboarding-class",
    ),
    re_path(
        rf"^teach/onboarding-class/(?P<access_code>{ACCESS_CODE_REGEX})/print_reminder_cards/$",
        teacher_print_reminder_cards,
        name="teacher_print_reminder_cards",
    ),
    re_path(
        rf"^teach/onboarding-class/(?P<access_code>{ACCESS_CODE_REGEX})/download_csv/$",
        teacher_download_csv,
        name="teacher_download_csv",
    ),
    re_path(
        r"^invited_teacher/(?P<token>[0-9a-f]+)/$",
        invited_teacher,
        name="invited_teacher",
    ),
    re_path(r"^play/$", play_landing_page, name="play"),
    re_path(
        r"^play/details/$",
        SchoolStudentDashboard.as_view(),
        name="student_details",
    ),
    re_path(
        r"^play/details/independent$",
        IndependentStudentDashboard.as_view(),
        name="independent_student_details",
    ),
    re_path(r"^play/account/$", student_edit_account, name="student_edit_account"),
    re_path(
        r"^play/account/independent/$",
        ratelimit(
            group=RATELIMIT_LOGIN_GROUP,
            key="post:name",
            method=RATELIMIT_METHOD,
            rate=RATELIMIT_LOGIN_RATE,
            block=True,
            is_teacher=False,
        )(independentStudentEditAccountView),
        name="independent_edit_account",
    ),
    re_path(
        r"^play/account/school_student/$",
        SchoolStudentEditAccountView.as_view(),
        name="school_student_edit_account",
    ),
    re_path(
        r"^play/join/$",
        student_join_organisation,
        name="student_join_organisation",
    ),
    re_path(r"^about", about, name="about"),
    re_path(r"^getinvolved", getinvolved, name="getinvolved"),
    re_path(r"^contribute", contribute, name="contribute"),
    re_path(r"^terms", terms, name="terms"),
    re_path(r"^privacy-notice/$", privacy_notice, name="privacy_notice"),
    re_path(r"^privacy-policy/$", privacy_notice, name="privacy_policy"),  # Keeping this to route from old URL
    re_path(r"^teach/dashboard/$", dashboard_manage, name="dashboard"),
    re_path(
        r"^teach/dashboard/kick/(?P<pk>[0-9]+)/$",
        organisation_kick,
        name="organisation_kick",
    ),
    re_path(
        r"^teach/dashboard/toggle_admin/(?P<pk>[0-9]+)/$",
        organisation_toggle_admin,
        name="organisation_toggle_admin",
    ),
    re_path(
        r"^teach/dashboard/disable_2FA/(?P<pk>[0-9]+)/$",
        teacher_disable_2FA,
        name="teacher_disable_2FA",
    ),
    re_path(
        r"^teach/dashboard/school/leave/$",
        organisation_leave,
        name="organisation_leave",
    ),
    re_path(
        r"^teach/dashboard/student/accept/(?P<pk>[0-9]+)/$",
        teacher_accept_student_request,
        name="teacher_accept_student_request",
    ),
    re_path(
        r"^teach/dashboard/student/reject/(?P<pk>[0-9]+)/$",
        teacher_reject_student_request,
        name="teacher_reject_student_request",
    ),
    re_path(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})$",
        teacher_view_class,
        name="view_class",
    ),
    re_path(
        rf"^teach/class/delete/(?P<access_code>{ACCESS_CODE_REGEX})$",
        teacher_delete_class,
        name="teacher_delete_class",
    ),
    re_path(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/delete/$",
        teacher_delete_students,
        name="teacher_delete_students",
    ),
    re_path(
        rf"^teach/class/edit/(?P<access_code>{ACCESS_CODE_REGEX})$",
        teacher_edit_class,
        name="teacher_edit_class",
    ),
    re_path(
        r"^teach/class/student/edit/(?P<pk>[0-9]+)/$",
        teacher_edit_student,
        name="teacher_edit_student",
    ),
    re_path(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/password_reset/$",
        teacher_class_password_reset,
        name="teacher_class_password_reset",
    ),
    re_path(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/dismiss/$",
        teacher_dismiss_students,
        name="teacher_dismiss_students",
    ),
    re_path(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/move/$",
        teacher_move_students,
        name="teacher_move_students",
    ),
    re_path(
        r"^teach/dashboard/resend_invite/(?P<token>[0-9a-f]+)/$",
        resend_invite_teacher,
        name="resend_invite_teacher",
    ),
    re_path(
        r"^teach/dashboard/toggle_admin_invite/(?P<invite_id>[0-9]+)/$",
        invite_toggle_admin,
        name="invite_toggle_admin",
    ),
    re_path(
        r"^teach/dashboard/delete_teacher_invite/(?P<token>[0-9a-f]+)$",
        delete_teacher_invite,
        name="delete_teacher_invite",
    ),
    re_path(
        rf"^teach/class/(?P<access_code>{ACCESS_CODE_REGEX})/students/move/disambiguate/$",
        teacher_move_students_to_class,
        name="teacher_move_students_to_class",
    ),
    re_path(r"^delete/account/$", delete_account, name="delete_account"),
    re_path(
        r"^schools/anonymise/(?P<start_id>\d+)/",
        AnonymiseOrphanSchoolsView.as_view(),
        name="anonymise_orphan_schools",
    ),
    re_path(
        r"^api/",
        include(
            [
                re_path(
                    r"^registered/(?P<year>\d{4})/(?P<month>\d{2})/(?P<day>\d{2})/$",
                    registered_users,
                    name="registered-users",
                ),
                re_path(
                    r"^lastconnectedsince/(?P<year>\d{4})/(?P<month>\d{2})/(?P<day>\d{2})/$",
                    last_connected_since,
                    name="last-connected-since",
                ),
                re_path(
                    r"^userspercountry/(?P<country>(AF|AX|AL|DZ|AS|AD|AO|AI|AQ|AG|AR|AM|AW|AU|AT|AZ|BS|BH|BD|BB|BY|BE|BZ|BJ|BM|BT|BO|BQ|BA|BW|BV|BR|IO|BN|BG|BF|BI|KH|CM|CA|CV|KY|CF|TD|CL|CN|CX|CC|CO|KM|CG|CD|CK|CR|CI|HR|CU|CW|CY|CZ|DK|DJ|DM|DO|EC|EG|SV|GQ|ER|EE|ET|FK|FO|FJ|FI|FR|GF|PF|TF|GA|GM|GE|DE|GH|GI|GR|GL|GD|GP|GU|GT|GG|GN|GW|GY|HT|HM|VA|HN|HK|HU|IS|IN|ID|IR|IQ|IE|IM|IL|IT|JM|JP|JE|JO|KZ|KE|KI|KP|KR|KW|KG|LA|LV|LB|LS|LR|LY|LI|LT|LU|MO|MK|MG|MW|MY|MV|ML|MT|MH|MQ|MR|MU|YT|MX|FM|MD|MC|MN|ME|MS|MA|MZ|MM|NA|NR|NP|NL|NC|NZ|NI|NE|NG|NU|NF|MP|NO|OM|PK|PW|PS|PA|PG|PY|PE|PH|PN|PL|PT|PR|QA|RE|RO|RU|RW|BL|SH|KN|LC|MF|PM|VC|WS|SM|ST|SA|SN|RS|SC|SL|SG|SX|SK|SI|SB|SO|ZA|GS|SS|ES|LK|SD|SR|SJ|SZ|SE|CH|SY|TW|TJ|TZ|TH|TL|TG|TK|TO|TT|TN|TR|TM|TC|TV|UG|UA|AE|GB|US|UM|UY|UZ|VU|VE|VN|VG|VI|WF|EH|YE|ZM|ZW))/$",
                    number_users_per_country,
                    name="number_users_per_country",
                ),
            ]
        ),
    ),
    re_path(r"^codingClub/$", coding_club, name="codingClub"),
    re_path(
        r"^removeFakeAccounts/",
        RemoveFakeAccounts.as_view(),
        name="remove_fake_accounts",
    ),
    re_path(r"^celebrate/", ten_year_map_page, name="celebrate"),
    re_path(
        r"^maintenance/$",
        TemplateView.as_view(template_name="maintenance.html"),
        name="maintenance",
    ),
]



from common.tests.utils.classes import generate_details


def create_class(page, teacher_id=None):
    name, _ = generate_details()

    page = page.create_class(name, "False", teacher_id=teacher_id)

    return page, name



from common import app_settings as common_app_settings
from common.utils import using_two_factor
from django import template
from django.conf import settings
from django.template.defaultfilters import stringfilter

from portal import __version__, beta

register = template.Library()


@register.filter(name="emaildomain")
@stringfilter
def emaildomain(email):
    return "*********" + email[email.find("@") :]


@register.filter(name="has_2FA")
def has_2FA(user):
    return using_two_factor(user)


@register.filter(name="is_logged_in")
def is_logged_in(user):
    return (
        user
        and user.is_authenticated
        and (
            not using_two_factor(user)
            or (hasattr(user, "is_verified") and user.userprofile.is_verified)
        )
    )


@register.filter
def is_developer(user):
    return not user.is_anonymous and user.userprofile.developer


@register.filter
def is_production(request):
    # 'production' excludes localhost/dev/staging
    return common_app_settings.MODULE_NAME == "default"


@register.filter
def has_beta_access(request):
    return beta.has_beta_access(request)


@register.filter(name="make_into_username")
def make_into_username(user):
    username = ""
    if hasattr(user, "userprofile"):
        if hasattr(user.userprofile, "student"):
            username = user.first_name
        elif hasattr(user.userprofile, "teacher"):
            username = user.first_name + " " + user.last_name

    return username


@register.filter(name="is_logged_in_as_teacher")
def is_logged_in_as_teacher(user):
    return (
        is_logged_in(user)
        and user.userprofile
        and hasattr(user.userprofile, "teacher")
    )


@register.filter(name="is_logged_in_as_admin_teacher")
def is_logged_in_as_admin_teacher(user):
    return is_logged_in_as_teacher(user) and user.userprofile.teacher.is_admin


@register.filter(name="is_logged_in_as_student")
def is_logged_in_as_student(user):
    return (
        is_logged_in(user)
        and user.userprofile
        and hasattr(user.userprofile, "student")
        and user.userprofile.student.class_field is not None
    )


@register.filter(name="is_independent_student")
def is_independent_student(user):
    return (
        is_logged_in(user)
        and user.userprofile
        and hasattr(user.userprofile, "student")
        and user.userprofile.student.is_independent()
    )


@register.filter(name="has_teacher_finished_onboarding")
def has_teacher_finished_onboarding(user):
    teacher = user.userprofile.teacher
    return is_logged_in_as_teacher(user) and teacher.has_school()


@register.filter(name="is_logged_in_as_school_user")
def is_logged_in_as_school_user(user):
    return (
        is_logged_in(user)
        and user.userprofile
        and (
            (
                hasattr(user.userprofile, "student")
                and user.userprofile.student.class_field is not None
            )
            or hasattr(user.userprofile, "teacher")
        )
    )


@register.filter(name="get_user_status")
def get_user_status(user):
    if is_logged_in_as_school_user(user):
        if is_logged_in_as_teacher(user):
            return "TEACHER"
        else:
            return "SCHOOL_STUDENT"
    elif is_logged_in(user):
        return "INDEPENDENT_STUDENT"
    else:
        return "UNTRACKED"


@register.filter(name="make_title_caps")
def make_title_caps(string):
    if len(string) > 0:
        string = string[0].upper() + string[1:]
    return string


@register.filter(name="cloud_storage")
@stringfilter
def cloud_storage(e):
    return settings.CLOUD_STORAGE_PREFIX + e


@register.filter(name="get_project_version")
def get_project_version():
    return __version__


@register.filter
def get_dict_item(dictionary, key):
    return dictionary[key]



from django.shortcuts import redirect

from portal.app_settings import TMP_AUTH_TOKEN


class TempBasicAuthMiddleware:
    """
    Middleware used as a basic auth for the new OTP app until we switch to the main domain.
    Only meant for use in OTP, so include this in MIDDLEWARES only in OTP settings, not GCP.
    Remove this middleware once the switch is done.
    """

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed on each request.
        url = request.build_absolute_uri()
        temp_domain = "https://www.code4life.education"

        # If the user inputs the token in the URL correctly, set a cookie to the token's value
        if url == f"{temp_domain}/{TMP_AUTH_TOKEN}":
            response = redirect(temp_domain)
            response.set_cookie(
                key="TMP_AUTH_TOKEN",
                value=TMP_AUTH_TOKEN,
            )
            return response
        # For any other request in the code4life domain, check that a cookie with the token's value exists.
        # Redirect to the main site if it doesn't.
        else:
            value = request.COOKIES.get("TMP_AUTH_TOKEN")
            if value is not None and value == TMP_AUTH_TOKEN:
                return self.get_response(request)

        return redirect("https://www.codeforlife.education")



from django import forms
from django.test import TestCase
from django_recaptcha.fields import ReCaptchaField
from django_recaptcha.widgets import ReCaptchaV2Invisible

from portal.helpers.captcha import is_captcha_in_form, remove_captcha_from_forms


class FormCaptchaTest(TestCase):
    class FormWithCaptcha(forms.Form):
        captcha = ReCaptchaField(widget=ReCaptchaV2Invisible)

    def test_is_captcha_in_form(self):
        form_with_captcha = FormCaptchaTest.FormWithCaptcha()
        print(form_with_captcha.fields)
        self.assertTrue(is_captcha_in_form(form_with_captcha))

        form_without_captcha = forms.Form()
        self.assertFalse(is_captcha_in_form(form_without_captcha))

    def test_remove_captcha_from_forms(self):
        form1 = FormCaptchaTest.FormWithCaptcha()
        form2 = forms.Form()
        form3 = FormCaptchaTest.FormWithCaptcha()

        remove_captcha_from_forms(form1, form2, form3)

        for form in [form1, form2, form3]:
            self.assertFalse(is_captcha_in_form(form))



# -*- coding: utf-8 -*-
import json
import re
import sys
import typing as t

from setuptools import find_packages, setup

with open("portal/__init__.py", "r") as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)


def parse_requirements(packages: t.Dict[str, t.Dict[str, t.Any]]):
    """Parse a group of requirements from `Pipfile.lock`.

    https://setuptools.pypa.io/en/latest/userguide/dependency_management.html

    Args:
        packages: The group name of the requirements.

    Returns:
        The requirements as a list of strings, required by `setuptools.setup`
    """

    requirements: t.List[str] = []
    for name, package in packages.items():
        requirement = name
        if requirement == "cfl-common":
            requirement += f"=={version}"
        else:
            if "extras" in package:
                requirement += f"[{','.join(package['extras'])}]"
            if "version" in package:
                requirement += package["version"]
            if "markers" in package:
                requirement += f"; {package['markers']}"
        requirements.append(requirement)

    return requirements


# Parse Pipfile.lock into strings.
with open("Pipfile.lock", "r", encoding="utf-8") as pipfile_lock:
    lock = json.load(pipfile_lock)
    install_requires = parse_requirements(lock["default"])
    dev_requires = parse_requirements(lock["develop"])

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="codeforlife-portal",
    version=version,
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={"dev": dev_requires},
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.12",
        "Framework :: Django",
    ],
    zip_safe=False,
)



from common.app_settings import MODULE_NAME
from common.utils import using_two_factor
from django.http import HttpResponseRedirect
from django.urls import reverse, reverse_lazy


class AdminAccessMiddleware(object):
    """
    This middleware only allows access to the Django admin site to users who are
    marked as superusers and who have 2FA enabled.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, callback, callback_args, callback_kwargs):
        if request.path.startswith(reverse("admin:index")):
            if request.user.is_authenticated:
                if not self._has_admin_access(request):
                    return HttpResponseRedirect(reverse_lazy("dashboard"))
            else:
                return HttpResponseRedirect(reverse_lazy("teacher_login"))

    def _has_admin_access(self, request):
        is_local = MODULE_NAME == "local"
        return (
            request.user.is_superuser
            and request.user.is_staff
            and (using_two_factor(request.user) or is_local)
        )
